#cython: embedsignature=True

# malloc and friends
from libc.stdlib cimport malloc, free
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair
from libc.stdio cimport printf

from cython.operator cimport dereference as deref, preincrement as inc

from .nnps_base cimport *
# Cython for compiler directives
cimport cython

import numpy as np
cimport numpy as np

DEF EPS = 1e-13

cdef extern from "<algorithm>" namespace "std" nogil:
    void sort[Iter, Compare](Iter first, Iter last, Compare comp)
    void sort[Iter](Iter first, Iter last)

IF UNAME_SYSNAME == "Windows":
    cdef inline double fmin(double x, double y) nogil:
        return x if x < y else y
    cdef inline double fmax(double x, double y) nogil:
        return x if x > y else y

    @cython.cdivision(True)
    cdef inline double log2(double n) nogil:
        return log(n)/log(2)

#############################################################################
cdef class StratifiedSFCNNPS(NNPS):

    """Finds nearest neighbors using Space-filling curves with stratified grids"""

    def __init__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int num_levels = 1,
            bint asymmetric = True):
        NNPS.__init__(
            self, dim, particles, radius_scale, ghost_layers, domain,
            cache, sort_gids
        )

        self.asymmetric = asymmetric

        self.radius_scale2 = radius_scale*radius_scale
        self.interval_size = 0

        self.src_index = 0
        self.dst_index = 0
        self.sort_gids = sort_gids
        self.domain.update()
        self.update()

    def __cinit__(self, int dim, list particles, double radius_scale = 2.0,
            int ghost_layers = 1, domain=None, bint fixed_h = False,
            bint cache = False, bint sort_gids = False, int num_levels = 1):

        cdef int narrays = len(particles)

        if fixed_h:
            self.num_levels = 1
        else:
            self.num_levels = num_levels

        self.pids = <uint32_t**> malloc(narrays*sizeof(uint32_t*))
        self.keys = <uint64_t**> malloc(narrays*sizeof(uint64_t*))
        self.key_to_idx = <int***> malloc(narrays*sizeof(int**))
        self.cell_sizes = <double**> malloc(narrays*sizeof(double*))
        self.hmax = <double***> malloc(narrays*sizeof(double**))
        self.max_keys = <uint64_t*> malloc(narrays*sizeof(uint64_t))
        self.num_cells = <int**> malloc(narrays*sizeof(int*))
        self.nbr_boxes = <vector[int]**> malloc(narrays*sizeof(void*))
        self.key_to_nbr_idx = <int***> malloc(narrays*sizeof(int**))
        self.key_to_nbr_length = <int***> malloc(narrays*sizeof(int**))
        self.total_mask_len = <int*> malloc(narrays*sizeof(int))

        cdef int i, j, num_particles
        for i from 0<=i<narrays:
            self.pids[i] = NULL
            self.keys[i] = NULL
            self.nbr_boxes[i] = NULL
            self.key_to_idx[i] = <int**> malloc(self.num_levels*sizeof(int*))
            self.key_to_nbr_idx[i] = <int**> malloc(self.num_levels*sizeof(int*))
            self.key_to_nbr_length[i] = <int**> malloc(self.num_levels*sizeof(int*))
            for j from 0<=j<self.num_levels:
                self.key_to_idx[i][j] = NULL
                self.key_to_nbr_idx[i][j] = NULL
                self.key_to_nbr_length[i][j] = NULL
            self.cell_sizes[i] = <double*> malloc(self.num_levels*sizeof(double))
            self.hmax[i] = <double**> malloc(self.num_levels*sizeof(double*))
            self.num_cells[i] = <int*> malloc(self.num_levels*sizeof(int))
            self.total_mask_len[i] = 0

        self.current_pids = NULL
        self.current_keys = NULL
        self.current_key_to_idx = NULL
        self.current_cells = NULL
        self.current_hmax = NULL
        self.current_nbr_boxes = NULL
        self.current_num_cells = NULL
        self.current_key_to_nbr_idx = NULL
        self.current_key_to_nbr_length = NULL
        self.current_mask_len = 0

    def __dealloc__(self):
        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int** current_key_to_idx
        cdef int** current_key_to_nbr_idx
        cdef int** current_key_to_nbr_length
        cdef double** current_hmax
        cdef int i, j
        for i from 0<=i<self.narrays:
            current_pids = self.pids[i]
            current_keys = self.keys[i]
            current_key_to_idx = self.key_to_idx[i]
            current_key_to_nbr_idx = self.key_to_nbr_idx[i]
            current_key_to_nbr_length = self.key_to_nbr_length[i]
            current_hmax = self.hmax[i]
            del self.nbr_boxes[i]
            for j from 0<=j<self.num_levels:
                free(current_key_to_idx[j])
                free(current_key_to_nbr_idx[j])
                free(current_key_to_nbr_length[j])
                free(current_hmax[j])
            free(current_pids)
            free(current_keys)
            free(current_key_to_idx)
            free(current_key_to_nbr_idx)
            free(current_key_to_nbr_length)
            free(current_hmax)
            free(self.cell_sizes[i])
            free(self.num_cells[i])
        free(self.pids)
        free(self.keys)
        free(self.key_to_idx)
        free(self.cell_sizes)
        free(self.key_to_nbr_idx)
        free(self.key_to_nbr_length)
        free(self.nbr_boxes)
        free(self.hmax)
        free(self.total_mask_len)
        free(self.max_keys)

    #### Public protocol ################################################

    cpdef double get_binning_size(self, int interval):
        """Get bin size at a level"""
        return self.radius_scale*self.current_cells[interval]

    cpdef set_context(self, int src_index, int dst_index):
        """Set context for nearest neighbor searches.

        Parameters
        ----------
        src_index: int
            Index in the list of particle arrays to which the neighbors belong

        dst_index: int
            Index in the list of particle arrays to which the query point belongs

        """
        NNPS.set_context(self, src_index, dst_index)
        self.current_pids = self.pids[src_index]
        self.current_keys = self.keys[src_index]
        self.current_cells = self.cell_sizes[src_index]
        self.current_key_to_idx = self.key_to_idx[src_index]
        self.current_key_to_nbr_idx = self.key_to_nbr_idx[src_index]
        self.current_key_to_nbr_length = self.key_to_nbr_length[src_index]
        self.current_max_key = self.max_keys[src_index]
        self.current_nbr_boxes = self.nbr_boxes[src_index]

        self.dst = <NNPSParticleArrayWrapper> self.pa_wrappers[dst_index]
        self.src = <NNPSParticleArrayWrapper> self.pa_wrappers[src_index]

    cpdef get_spatially_ordered_indices(self, int pa_index, LongArray indices):
        indices.reset()
        cdef int num_particles = (<NNPSParticleArrayWrapper> \
                self.pa_wrappers[pa_index]).get_number_of_particles()
        cdef uint32_t* current_pids = self.pids[pa_index]

        cdef int j
        for j from 0<=j<num_particles:
            indices.c_append(current_pids[j])

    @cython.cdivision(True)
    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil:
        """Low level, high-performance non-gil method to find neighbors.
        This requires that `set_context()` be called beforehand.  This method
        does not reset the neighbors array before it appends the
        neighbors to it.

        """
        cdef double* dst_x_ptr = self.dst.x.data
        cdef double* dst_y_ptr = self.dst.y.data
        cdef double* dst_z_ptr = self.dst.z.data
        cdef double* dst_h_ptr = self.dst.h.data

        cdef double* src_x_ptr = self.src.x.data
        cdef double* src_y_ptr = self.src.y.data
        cdef double* src_z_ptr = self.src.z.data
        cdef double* src_h_ptr = self.src.h.data

        cdef double x = dst_x_ptr[d_idx]
        cdef double y = dst_y_ptr[d_idx]
        cdef double z = dst_z_ptr[d_idx]
        cdef double h = dst_h_ptr[d_idx]

        cdef int num_particles = self.src.x.length

        cdef unsigned int* s_gid = self.src.gid.data
        cdef int orig_length = nbrs.length

        cdef int c_x, c_y, c_z
        cdef double* xmin = self.xmin.data
        cdef unsigned int i, pid
        cdef int idx

        cdef double xij2 = 0
        cdef double hi2 = self.radius_scale2 * h * h
        cdef double hj2 = 0

        cdef uint64_t key

        cdef uint32_t q_level = self._get_level(h)

        find_cell_id_raw(
                x - xmin[0],
                y - xmin[1],
                z - xmin[2],
                self.radius_scale*self.current_cells[q_level],
                &c_x, &c_y, &c_z
                )

        cdef int* current_key_to_nbr_idx_level
        cdef int* current_key_to_nbr_length_level
        cdef int num_boxes, start_idx

        current_key_to_nbr_idx_level = self.current_key_to_nbr_idx[q_level]
        current_key_to_nbr_length_level = self.current_key_to_nbr_length[q_level]

        find_cell_id_raw(
                x - xmin[0],
                y - xmin[1],
                z - xmin[2],
                self.radius_scale*self.current_cells[0],
                &c_x, &c_y, &c_z
                )

        key_bottom = get_key(c_x, c_y, c_z)

        num_boxes = current_key_to_nbr_length_level[key_bottom]
        start_idx = current_key_to_nbr_idx_level[key_bottom]

        if not (start_idx >= 0 and num_boxes > 0):
            return

        for i from start_idx<=i<start_idx + num_boxes:
            idx = deref(self.current_nbr_boxes)[i]
            key = self.current_keys[idx]

            while idx < num_particles and self.current_keys[idx] == key:
                pid = self.current_pids[idx]

                hj2 = self.radius_scale2*src_h_ptr[pid]*src_h_ptr[pid]

                xij2 = norm2(
                    src_x_ptr[pid] - x,
                    src_y_ptr[pid] - y,
                    src_z_ptr[pid] - z
                    )

                if (xij2 < hi2) or (xij2 < hj2):
                    nbrs.c_append(pid)

                idx += 1

        if self.sort_gids:
            self._sort_neighbors(
                &nbrs.data[orig_length], nbrs.length - orig_length, s_gid
            )

    cpdef int get_number_of_particles(self, int pa_index, int level):
        cdef int length = 0
        cdef int i = 0
        cdef int num_particles = (<NNPSParticleArrayWrapper> \
                self.pa_wrappers[pa_index]).get_number_of_particles()
        cdef uint64_t* current_keys = self.keys[pa_index]

        while i < num_particles and (current_keys[i] >> self.max_num_bits) != level:
            i += 1

        if i == num_particles:
            return length

        length += 1
        i += 1

        while i < num_particles and (current_keys[i] >> self.max_num_bits == \
                current_keys[i-1] >> self.max_num_bits):
                    length += 1
                    i += 1

        return length

    #### Private protocol ################################################
    cpdef np.ndarray get_keys(self, pa_index):
        cdef uint64_t* current_keys = self.keys[pa_index]
        cdef int num_particles = (<NNPSParticleArrayWrapper> \
                self.pa_wrappers[pa_index]).get_number_of_particles()
        keys = np.empty(num_particles)
        for i from 0<=i<num_particles:
            keys[i] = current_keys[i]
        return keys

    @cython.cdivision(True)
    cdef inline int _get_level(self, double h) nogil:
        return self.num_levels - <int> min(self.num_levels,
                ceil(log2((self.cell_size + EPS)/ self.radius_scale / h)))

    @cython.cdivision(True)
    cdef inline int _get_H(self, double h_q, double h_j):
        return <int> ceil(h_q / h_j)

    cdef inline int get_idx(self, uint64_t key, uint64_t max_key, int* key_to_idx) nogil:
        return -1 if key >= max_key else key_to_idx[key]

    cdef int _neighbor_boxes_func(self, int i, int j, int k, int H,
            int* current_key_to_idx_level, uint64_t max_key,
            double current_cell_size, double* current_hmax_level, 
            vector[int]* nbr_boxes):
        if self.asymmetric:
            return self._neighbor_boxes_asym(i, j, k, H, current_key_to_idx_level,
                    max_key, current_cell_size, current_hmax_level, nbr_boxes)
        else:
            return self._neighbor_boxes_sym(i, j, k, H, current_key_to_idx_level,
                    max_key, current_cell_size, current_hmax_level, nbr_boxes)

    cdef int _neighbor_boxes_asym(self, int i, int j, int k, int H,
            int* current_key_to_idx_level, uint64_t max_key,
            double current_cell_size, double* current_hmax_level,
            vector[int]* nbr_boxes) nogil:
        cdef int length = 0

        cdef uint64_t key
        cdef int found_idx

        cdef int x_temp, y_temp, z_temp

        cdef int s, t, u

        for s from -H<=s<=H:
            for t from -H<=t<=H:
                for u from -H<=u<=H:

                    x_temp = i + u
                    y_temp = j + t
                    z_temp = k + s

                    if x_temp >= 0 and y_temp >= 0 and z_temp >= 0:
                        key = get_key(x_temp, y_temp, z_temp)
                        found_idx = self.get_idx(key, max_key, current_key_to_idx_level)

                        if found_idx == -1:
                            continue

                        nbr_boxes.push_back(found_idx)
                        length += 1

        return length

    @cython.cdivision(True)
    cdef int _neighbor_boxes_sym(self, int i, int j, int k, int H,
            int* current_key_to_idx_level, uint64_t max_key,
            double current_cell_size, double* current_hmax_level,
            vector[int]* nbr_boxes) nogil:
        cdef int length = 0

        cdef uint64_t key

        cdef int x_temp, y_temp, z_temp

        cdef double h_local
        cdef int H_eff

        cdef int s, t, u

        cdef uint64_t qkey = get_key(i, j, k)

        for s from -H<=s<=H:
            for t from -H<=t<=H:
                for u from -H<=u<=H:

                    x_temp = i + u
                    y_temp = j + t
                    z_temp = k + s

                    if x_temp >= 0 and y_temp >= 0 and z_temp >= 0:
                        key = get_key(x_temp, y_temp, z_temp)
                        found_idx = self.get_idx(key, max_key, current_key_to_idx_level)

                        if found_idx == -1:
                            continue

                        h_local = fmax(current_hmax_level[key], current_hmax_level[qkey])
                        H_eff = <int> ceil(h_local / current_cell_size)

                        if abs(u) <= H_eff and abs(t) <= H_eff and abs(s) <= H_eff:
                            nbr_boxes.push_back(found_idx)
                            length += 1

        return length

    cdef void _fill_nbr_boxes(self):
        cdef int i, j, k
        cdef NNPSParticleArrayWrapper pa_wrapper
        cdef int num_particles

        cdef vector[int]* current_nbr_boxes
        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int** current_key_to_idx
        cdef double** current_hmax

        cdef double* x_ptr
        cdef double* y_ptr
        cdef double* z_ptr
        cdef double* h_ptr

        cdef int c_x, c_y, c_z
        cdef int num_boxes

        cdef int n = 0
        cdef uint32_t pid

        cdef uint64_t key, key_bottom
        cdef uint64_t key_stripped, strip_mask, current_max_key
        cdef int mask_length = 0

        strip_mask = (1 << self.max_num_bits) - 1

        for i from 0<=i<self.narrays:
            n = 0
            pa_wrapper = self.pa_wrappers[i]
            num_particles = pa_wrapper.get_number_of_particles()
            x_ptr = pa_wrapper.x.data
            y_ptr = pa_wrapper.y.data
            z_ptr = pa_wrapper.z.data
            h_ptr = pa_wrapper.h.data
            xmin = self.xmin.data

            current_keys = self.keys[i]
            current_key_to_idx = self.key_to_idx[i]
            current_pids = self.pids[i]
            current_nbr_boxes = self.nbr_boxes[i]
            current_cells = self.cell_sizes[i]
            current_hmax = self.hmax[i]
            current_key_to_nbr_idx = self.key_to_nbr_idx[i]
            current_key_to_nbr_length = self.key_to_nbr_length[i]
            current_max_key = self.max_keys[i]

            # init self.nbr_boxes to -1
            for j from 0<=j<self.total_mask_len[i]:
                deref(current_nbr_boxes)[j] = -1

            for j from 0<=j<num_particles:
                key = current_keys[j]
                pid = current_pids[j]
                level = key >> self.max_num_bits
                key_stripped = key & strip_mask
                current_hmax_level = current_hmax[level]
                if j == 0 or key != current_keys[j-1]:
                    current_hmax_level[key_stripped] = h_ptr[pid]
                else:
                    current_hmax_level[key_stripped] = fmax(current_hmax_level[key_stripped], h_ptr[pid])
            
            pid = current_pids[0]
            key = current_keys[0]
            
            find_cell_id_raw(
                x_ptr[pid] - xmin[0],
                y_ptr[pid] - xmin[1],
                z_ptr[pid] - xmin[2],
                self.radius_scale*current_cells[0],
                &c_x, &c_y, &c_z
                )

            key_bottom = get_key(c_x, c_y, c_z)

            level = key >> self.max_num_bits
            key_stripped = key & strip_mask
            current_hmax_level = current_hmax[level]
            hmax_cell = current_hmax_level[key_stripped]
            mask_length = 0
            
            current_key_to_nbr_idx_level = current_key_to_nbr_idx[level]
            current_key_to_nbr_length_level = current_key_to_nbr_length[level]

            current_key_to_nbr_idx_level[key_bottom] = n

            for k from 0<=k<self.num_levels:
                if current_cells[k] == 0:
                    continue

                find_cell_id_raw(
                    x_ptr[pid] - xmin[0],
                    y_ptr[pid] - xmin[1],
                    z_ptr[pid] - xmin[2],
                    self.radius_scale*current_cells[k],
                    &c_x, &c_y, &c_z
                    )

                H = self._get_H(hmax_cell, current_cells[k])
                
                num_boxes = self._neighbor_boxes_func(
                        c_x, c_y, c_z,
                        H, current_key_to_idx[k], current_max_key,
                        current_cells[k], current_hmax[k],
                        current_nbr_boxes
                        )

                n += num_boxes
                mask_length += num_boxes
            
            current_key_to_nbr_length_level[key_bottom] = mask_length

            # nbrs of all cids in particle array i
            for j from 0<j<num_particles:
                key = current_keys[j]
                pid = current_pids[j]

                find_cell_id_raw(
                    x_ptr[pid] - xmin[0],
                    y_ptr[pid] - xmin[1],
                    z_ptr[pid] - xmin[2],
                    self.radius_scale*current_cells[0],
                    &c_x, &c_y, &c_z
                    )

                key_bottom = get_key(c_x, c_y, c_z)

                level = key >> self.max_num_bits
                key_stripped = key & strip_mask
                current_hmax_level = current_hmax[level]
                hmax_cell = current_hmax_level[key_stripped]
                mask_length = 0
                current_key_to_nbr_idx_level = current_key_to_nbr_idx[level]
                current_key_to_nbr_length_level = current_key_to_nbr_length[level]
                    
                if current_key_to_nbr_idx_level[key_bottom] == -1:
                    current_key_to_nbr_idx_level[key_bottom] = n
 
                    for k from 0<=k<self.num_levels:
                        if current_cells[k] == 0:
                            continue

                        find_cell_id_raw(
                            x_ptr[pid] - xmin[0],
                            y_ptr[pid] - xmin[1],
                            z_ptr[pid] - xmin[2],
                            self.radius_scale*current_cells[k],
                            &c_x, &c_y, &c_z
                            )

                        H = self._get_H(hmax_cell, current_cells[k])
                        
                        num_boxes = self._neighbor_boxes_func(
                                c_x, c_y, c_z,
                                H, current_key_to_idx[k], current_max_key,
                                current_cells[k], current_hmax[k],
                                current_nbr_boxes
                                )

                        n += num_boxes
                        mask_length += num_boxes
                    
                    current_key_to_nbr_length_level[key_bottom] = mask_length

    @cython.cdivision(True)
    cpdef _refresh(self):
        self.interval_size = ((self.cell_size - self.hmin)/self.num_levels) + EPS

        cdef double* current_cells

        cdef int i, j, num_particles
        for i from 0<=i<self.narrays:
            num_particles = (<NNPSParticleArrayWrapper> \
                    self.pa_wrappers[i]).get_number_of_particles()
            if self.pids[i] != NULL:
                free(self.pids[i])
            if self.keys[i] != NULL:
                free(self.keys[i])
            if self.nbr_boxes[i] != NULL:
                del self.nbr_boxes[i]
            self.pids[i] = <uint32_t*> malloc(num_particles*sizeof(uint32_t))
            self.keys[i] = <uint64_t*> malloc(num_particles*sizeof(uint64_t))
            current_cells = self.cell_sizes[i]
            for j from 0<=j<self.num_levels:
                current_cells[j] = 0
                if self.key_to_idx[i][j] != NULL:
                    free(self.key_to_idx[i][j])
                if self.key_to_nbr_idx[i][j] != NULL:
                    free(self.key_to_nbr_idx[i][j])
                if self.key_to_nbr_length[i][j] != NULL:
                    free(self.key_to_nbr_length[i][j])
        self.current_pids = self.pids[self.src_index]
        self.current_cells = self.cell_sizes[self.src_index]

        cdef double max_length = fmax(fmax((self.xmax.data[0] - \
                self.xmin.data[0]),
            (self.xmax.data[1] - self.xmin.data[1])),
            (self.xmax.data[2] - self.xmin.data[2]))

        cdef int max_num_cells = (<int> ceil(max_length/self.hmin))

        self.max_num_bits = 1 + 3*(<int> ceil(log2(max_num_cells)))

        self.max_possible_key = 0

        cdef NNPSParticleArrayWrapper pa_wrapper

        cdef uint32_t* current_pids
        cdef uint64_t* current_keys
        cdef int** current_key_to_nbr_idx 
        cdef int** current_key_to_nbr_length
        cdef int* current_key_to_nbr_idx_level
        cdef int* current_key_to_nbr_length_level
        cdef uint64_t key_iter

        cdef int H_jk, num_cells_level

        for i from 0<=i<self.narrays:
            self.total_mask_len[i] = 0
            pa_wrapper = self.pa_wrappers[i]
            current_cells = self.cell_sizes[i]
            current_pids = self.pids[i]
            current_keys = self.keys[i]
            current_num_cells = self.num_cells[i]

            self.fill_array(pa_wrapper, i, current_pids,
                    current_keys, current_cells, current_num_cells)

            current_key_to_nbr_idx = self.key_to_nbr_idx[i]
            current_key_to_nbr_length = self.key_to_nbr_idx[i]
            
            for j from 0<=j<self.num_levels:
                num_cells_level = current_num_cells[j]
                for k from 0<=k<self.num_levels:
                    if current_cells[j] == 0 or current_cells[k] == 0:
                        continue
                    H_jk = self._get_H(current_cells[j], current_cells[k])
                    self.total_mask_len[i] += num_cells_level * (2 * H_jk + 1) ** 3

            self.nbr_boxes[i] = new vector[int]()
            self.nbr_boxes[i].reserve(self.total_mask_len[i])

        self.max_possible_key += 1

        for i from 0<=i<self.narrays:
            current_key_to_nbr_idx = self.key_to_nbr_idx[i]
            current_key_to_nbr_length = self.key_to_nbr_length[i]
            for j from 0<=j<self.num_levels:
                current_key_to_nbr_idx[j] = <int*> malloc(self.max_possible_key * sizeof(int))
                current_key_to_nbr_length[j] = <int*> malloc(self.max_possible_key * sizeof(int))
                current_key_to_nbr_idx_level = current_key_to_nbr_idx[j]
                current_key_to_nbr_length_level = current_key_to_nbr_length[j]

                for key_iter from 0<=key_iter<self.max_possible_key:
                    current_key_to_nbr_idx_level[key_iter] = -1
                    current_key_to_nbr_length_level[key_iter] = -1

        self._fill_nbr_boxes()

    @cython.cdivision(True)
    cdef void fill_array(self, NNPSParticleArrayWrapper pa_wrapper,
            int pa_index, uint32_t* current_pids,
            uint64_t* current_keys, double* current_cells, int* current_num_cells):
        cdef double* x_ptr = pa_wrapper.x.data
        cdef double* y_ptr = pa_wrapper.y.data
        cdef double* z_ptr = pa_wrapper.z.data
        cdef double* h_ptr = pa_wrapper.h.data

        cdef int curr_num_particles = pa_wrapper.get_number_of_particles()

        cdef double* xmax = self.xmax.data
        cdef double* xmin = self.xmin.data

        cdef int c_x, c_y, c_z

        cdef int i, j, level
        cdef uint64_t level_padded, key

        # Finds cell sizes at each level
        for i from 0<=i<self.num_levels:
            current_cells[i] = (self.cell_size / self.radius_scale) / (2 ** (self.num_levels - i - 1))

        cdef double cell_size

        cdef uint64_t max_key = 0

        for i from 0<=i<curr_num_particles:
            current_pids[i] = i
            level = self._get_level(h_ptr[i])
            level_padded = level << self.max_num_bits
            cell_size = self.radius_scale*current_cells[level]
            find_cell_id_raw(
                    x_ptr[i] - xmin[0],
                    y_ptr[i] - xmin[1],
                    z_ptr[i] - xmin[2],
                    cell_size,
                    &c_x, &c_y, &c_z
                    )
            key = get_key(c_x, c_y, c_z)
            current_keys[i] = level_padded + key

            max_key = max(max_key, key)

            find_cell_id_raw(
                    x_ptr[i] - xmin[0],
                    y_ptr[i] - xmin[1],
                    z_ptr[i] - xmin[2],
                    self.radius_scale*current_cells[0],
                    &c_x, &c_y, &c_z
                    )
            key = get_key(c_x, c_y, c_z)
            self.max_possible_key = max(key, self.max_possible_key)

        max_key += 1

        self.max_keys[pa_index] = max_key

        cdef CompareSortWrapper sort_wrapper = CompareSortWrapper(
                current_pids, current_keys, curr_num_particles)

        sort_wrapper.compare_sort()

        cdef int** current_key_to_idx = self.key_to_idx[pa_index]
        cdef double** current_hmax = self.hmax[pa_index]
        cdef int* key_idx_level
        cdef double* hmax_level

        for i from 0<=i<self.num_levels:
            current_key_to_idx[i] = <int*> malloc(max_key * sizeof(int))
            current_hmax[i] = <double*> malloc(max_key * sizeof(double))
            key_idx_level = current_key_to_idx[i]
            hmax_level = current_hmax[i]
            current_num_cells[i] = 0

            for j from 0<=j<max_key:
                key_idx_level[j] = -1
                hmax_level[j] = -1

        ################################################################

        cdef uint64_t key_stripped, strip_mask

        strip_mask = (1 << self.max_num_bits) - 1

        key = current_keys[0]
        level = key >> self.max_num_bits
        key_stripped = key & strip_mask
        current_key_to_idx[level][key_stripped] = 0;
        current_num_cells[level] += 1

        for i from 0<i<curr_num_particles:
            key = current_keys[i]
            if key != current_keys[i-1]:
                level = key >> self.max_num_bits
                key_idx_level = current_key_to_idx[level]
                key_stripped = key & strip_mask
                key_idx_level[key_stripped] = i
                current_num_cells[level] += 1

    @cython.cdivision(True)
    cpdef _bin(self, int pa_index, UIntArray indices):
        pass



from collections import defaultdict
from os.path import dirname, join, expanduser, realpath
from textwrap import dedent

from mako.template import Template
from cyarray import carray

from compyle.config import get_config
from compyle.cython_generator import (CythonGenerator, KnownType,
                                      get_parallel_range)
from compyle.ext_module import ExtModule, get_platform_dir


###############################################################################
def get_cython_code(obj):
    """This function looks at the object and gets any additional code to
    wrap from either the `_cython_code_` method or the `_get_helpers_` method.
    """
    result = []
    if hasattr(obj, '_cython_code_'):
        code = obj._cython_code_()
        doc = '# From %s' % obj.__class__.__name__
        result.extend([doc, code] if len(code) > 0 else [])
    return result


def get_helper_code(helpers):
    """Given a list of helpers, return the helper code suitably wrapped.
    """
    result = []
    result.append('# Helpers')
    cg = CythonGenerator()
    for helper in helpers:
        cg.parse(helper)
        result.append(cg.get_code())
    return result


###############################################################################
def get_all_array_names(particle_arrays):
    """For each type of carray, find the union of the names of all particle
    array properties/constants along with their array type.  Returns a
    dictionary keyed on the name of the Array class with values being a set of
    property names for each.

    Parameters
    ----------

    particle_array : list
        A list of particle arrays.

    Examples
    --------

    A simple example would be::

       >>> x = np.linspace(0, 1, 10)
       >>> pa = ParticleArray(name='f', x=x)
       >>> get_all_array_names([pa])
       {'DoubleArray': {'x'}, 'IntArray': {'pid', 'tag'}, 'UIntArray': {'gid'}}
    """
    props = defaultdict(set)
    for array in particle_arrays:
        for properties in (array.properties, array.constants):
            for name, arr in properties.items():
                a_type = arr.__class__.__name__
                props[a_type].add(name)
    return dict(props)


def get_known_types_for_arrays(array_names):
    """Given all the array names from `get_all_array_names` this creates known
    types for each of them so that the code generators can use this type
    information when needed.  Note that known type info is generated for both
    source and destination style arrays.

    Parameters
    ----------

    array_names: dict
        A dictionary produced by `get_all_array_names`.

    Examples
    --------

    A simple example would be::

        >>> x = np.linspace(0, 1, 10)
        >>> pa = ParticleArray(name='f', x=x)
        >>> pa.remove_property('pid')
        >>> info = get_all_array_names([pa])
        >>> get_known_types_for_arrays(info)
        {'d_gid': KnownType("unsigned int*"),
         'd_tag': KnownType("int*"),
         'd_x': KnownType("double*"),
         's_gid': KnownType("unsigned int*"),
         's_tag': KnownType("int*"),
         's_x': KnownType("double*")}

    """
    result = {}
    for arr_type, arrays in array_names.items():
        c_type = getattr(carray, arr_type)().get_c_type()
        for arr in arrays:
            known_type = KnownType(c_type + '*')
            result['s_' + arr] = known_type
            result['d_' + arr] = known_type
    return result


###############################################################################
class AccelerationEvalCythonHelper(object):
    def __init__(self, acceleration_eval):
        self.object = acceleration_eval
        self.config = get_config()
        self.all_array_names = get_all_array_names(
            self.object.particle_arrays
        )
        self.known_types = get_known_types_for_arrays(
            self.all_array_names
        )
        self._ext_mod = None
        self._module = None
        self._compute_group_map()

    ##########################################################################
    # Private interface.
    ##########################################################################
    def _compute_group_map(self):
        # Given all the groups, create a mapping from the group to an index of
        # sorts that can be used when adding the pre/post callback code.
        mapping = {}
        for g_idx, group in enumerate(self.object.mega_groups):
            mapping[group] = 'self.groups[%d]' % g_idx
            if group.has_subgroups:
                for sg_idx, sub_group in enumerate(group.data):
                    code = 'self.groups[{gid}].data[{sgid}]'.format(
                        gid=g_idx, sgid=sg_idx
                    )
                    mapping[sub_group] = code
        self._group_map = mapping

    ##########################################################################
    # Public interface.
    ##########################################################################
    def get_code(self):
        path = join(dirname(__file__), 'acceleration_eval_cython.mako')
        template = Template(filename=path)
        main = template.render(helper=self)
        return main

    def setup_compiled_module(self, module):
        # Create the compiled module.
        object = self.object
        acceleration_eval = module.AccelerationEval(
            object.kernel, object.all_group.equations,
            object.particle_arrays, object.mega_groups
        )
        object.set_compiled_object(acceleration_eval)

    def compile(self, code):
        # Note, we do not add carray or particle_array as nnps_base would
        # have been rebuilt anyway if they changed.
        root = expanduser(join('~', '.pysph', 'source', get_platform_dir()))
        depends = ["pysph.base.nnps_base"]
        # Add pysph/base directory to inc_dirs for including spatial_hash.h
        # for SpatialHashNNPS
        extra_inc_dirs = [join(dirname(dirname(realpath(__file__))), 'base')]
        self._ext_mod = ExtModule(
            code, verbose=False, root=root, depends=depends,
            extra_inc_dirs=extra_inc_dirs
        )
        self._module = self._ext_mod.load()
        return self._module

    ##########################################################################
    # Mako interface.
    ##########################################################################
    def get_array_decl_for_wrapper(self):
        array_names = self.all_array_names
        decl = []
        for a_type in sorted(array_names.keys()):
            props = array_names[a_type]
            decl.append(
                'cdef public {a_type} {attrs}'.format(
                    a_type=a_type, attrs=', '.join(sorted(props))
                )
            )
        return '\n'.join(decl)

    def get_header(self):
        object = self.object
        helpers = []
        headers = []
        headers.extend(get_cython_code(object.kernel))
        if hasattr(object.kernel, '_get_helpers_'):
            helpers.extend(object.kernel._get_helpers_())

        # get headers from the Equations
        for equation in object.all_group.equations:
            headers.extend(get_cython_code(equation))
            if hasattr(equation, '_get_helpers_'):
                for helper in equation._get_helpers_():
                    if helper not in helpers:
                        helpers.append(helper)

        headers.extend(get_helper_code(helpers))

        # Kernel wrappers.
        cg = CythonGenerator(known_types=self.known_types)
        cg.parse(object.kernel)
        headers.append(cg.get_code())

        # Equation wrappers.
        self.known_types['SPH_KERNEL'] = KnownType(
            object.kernel.__class__.__name__
        )
        headers.append(object.all_group.get_equation_wrappers(
            self.known_types
        ))

        return '\n'.join(headers)

    def get_equation_defs(self):
        return self.object.all_group.get_equation_defs()

    def get_equation_init(self):
        return self.object.all_group.get_equation_init()

    def get_kernel_defs(self):
        return 'cdef public %s kernel' % (
            self.object.kernel.__class__.__name__
        )

    def get_kernel_init(self):
        object = self.object
        return 'self.kernel = %s(**kernel.__dict__)' % (
            object.kernel.__class__.__name__
        )

    def get_variable_declarations(self):
        group = self.object.all_group
        ctx = group.context
        return group.get_variable_declarations(ctx)

    def get_array_declarations(self):
        group = self.object.all_group
        src, dest = group.get_array_names()
        src.update(dest)
        return group.get_array_declarations(src, self.known_types)

    def get_dest_array_setup(self, dest_name, eqs_with_no_source, sources,
                             group):
        src, dest_arrays = eqs_with_no_source.get_array_names()
        for g in sources.values():
            s, d = g.get_array_names()
            dest_arrays.update(d)
        if isinstance(group.start_idx, str):
            lines = ['D_START_IDX = self.%s.%s[0]' %
                     (dest_name, group.start_idx)]
        else:
            lines = ['D_START_IDX = %s' % group.start_idx]

        if group.stop_idx is None:
            lines += ['NP_DEST = self.%s.size(real=%s)' %
                      (dest_name, group.real)]
        elif isinstance(group.stop_idx, str):
            lines += ['NP_DEST = self.%s.%s[0]' %
                      (dest_name, group.stop_idx)]
        else:
            lines += ['NP_DEST = %s' % group.stop_idx]

        lines += ['%s = dst.%s.data' % (n, n[2:])
                  for n in sorted(dest_arrays)]
        return '\n'.join(lines)

    def get_src_array_setup(self, src_name, eq_group):
        src_arrays, dest = eq_group.get_array_names()
        lines = ['NP_SRC = self.%s.size()' % src_name]
        lines += ['%s = src.%s.data' % (n, n[2:])
                  for n in sorted(src_arrays)]
        return '\n'.join(lines)

    def get_parallel_block(self):
        if self.config.use_openmp:
            return "with nogil, parallel():"
        else:
            return "if True: # Placeholder used for OpenMP."

    def get_parallel_range(self, group, nogil=True):
        kwargs = {}
        if (group.stop_idx is not None) or group.start_idx:
            kwargs['schedule'] = 'dynamic'
            kwargs['chunksize'] = None
        if nogil:
            kwargs['nogil'] = True

        return get_parallel_range("D_START_IDX", "NP_DEST", **kwargs)

    def get_particle_array_names(self):
        parrays = [pa.name for pa in self.object.particle_arrays]
        return ', '.join(parrays)

    def get_condition_call(self, group):
        return self._group_map[group] + '.condition(t, dt)'

    def get_pre_call(self, group):
        return self._group_map[group] + '.pre()'

    def get_post_call(self, group):
        return self._group_map[group] + '.post()'

    def get_iteration_init(self, group):
        lines = [
            'max_iterations = %d' % group.max_iterations,
            'min_iterations = %d' % group.min_iterations,
            '_iteration_count = 1',
            'while True:'
        ]
        return '\n'.join(lines)

    def get_iteration_check(self, group):
        src = dedent('''\
            ###############################################################
            ## Break the iteration for the group.
            ###############################################################
            if ((_iteration_count >= min_iterations)
               and (%s or (_iteration_count == max_iterations))):
                _iteration_count = 1
                break
            _iteration_count += 1
        ''' % group.get_converged_condition())
        return src



"""High-velocity impact of an Steel projectile on an Aluminium plate"""

import numpy

# SPH equations
from pysph.sph.equation import Group

from pysph.sph.basic_equations import IsothermalEOS, ContinuityEquation, MonaghanArtificialViscosity,\
     XSPHCorrection, VelocityGradient3D

from pysph.sph.solid_mech.basic import MomentumEquationWithStress, HookesDeviatoricStressRate,\
    MonaghanArtificialStress, EnergyEquationWithStress

from pysph.sph.solid_mech.hvi import VonMisesPlasticity2D, MieGruneisenEOS, StiffenedGasEOS

from pysph.sph.gas_dynamics.basic import UpdateSmoothingLengthFromVolume

from pysph.base.utils import get_particle_array
from pysph.base.kernels import Gaussian, CubicSpline, WendlandQuintic
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.sph.integrator import PECIntegrator, EPECIntegrator
from pysph.sph.integrator_step import SolidMechStep

def add_properties(pa, *props):
    for prop in props:
        pa.add_property(name=prop)

# Parameters
dx = dy = dz = 0.0005 # m
hdx = 1.3
h = hdx*dx
r = 0.005

######################################################################
# Material properties: Table (1) of "A Free Lagrange Augmented Godunov Method
# for the Simulation of Elastic-Plastic Solids", B. P. Howell and G. J. Ball,
# JCP (2002)

# ALUMINIUM
ro1 = 2785.0         # refenrence density
C1 = 5328.0          # reference sound-speed
S1 = 1.338           # Particle-shock velocity slope
gamma1 = 2.0         # Gruneisen gamma/parameter
G1 = 2.76e7          # Shear Modulus (kPa)
Yo1 = 0.3e6          # Yield stress
E1 = ro1*C1*C1       # Youngs Modulus

# STEEL
ro2 = 7900.0         # reference density
C2 = 4600.0          # reference sound-speed
S2 = 1.490           # Particle shock velocity slope
gamma2 = 2.17        # Gruneisen gamma/parameter
G2 = 8.530e7         # Shear modulus
Yo2 = 0.979e6        # Yield stress
E2 = ro2*C2*C2       # Youngs modulus

# general
v_s = 3100.0         # Projectile velocity 3.1 km/s
cs1=numpy.sqrt(E1/ro1) # speed of sound in aluminium
cs2=numpy.sqrt(E2/ro2) # speed of sound in steel

######################################################################
# SPH constants and parameters

# Monaghan-type artificial viscosity
avisc_alpha = 1.0; avisc_beta = 1.5; avisc_eta = 0.1

# XSPH epsilon
xsph_eps = 0.5

# SAV1 artificial viscosity coefficients
alpha1 = 1.0
beta1 = 1.5
eta = 0.1# in piab equation eta2 was written so final value is 0.01.(as req.)

# SAV2
alpha2 = 2.5
beta2 = 2.5
eta = 0.1# in piab equation eta2 was written so final value is 0.01.(as req.)

# XSPH
eps = 0.5

######################################################################
# Particle creation rouintes
def get_projectile_particles():

    x, y, z = numpy.mgrid[-r:r+1e-6:dx, -r:r+1e-6:dx, -r:r+1e-6:dx]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()

    d = (x*x+y*y+z*z)
    keep = numpy.flatnonzero(d<=r*r)
    x = x[keep]
    y = y[keep]
    z = z[keep]

    x = x-(r+2*dx)
    print('%d Projectile particles'%len(x))

    hf = numpy.ones_like(x) * h
    mf = numpy.ones_like(x) * dx * dy * dz * ro2
    rhof = numpy.ones_like(x) * ro2
    csf  = numpy.ones_like(x) * cs2
    u    = numpy.ones_like(x) * v_s

    pa = projectile = get_particle_array(
        name="projectile", x=x, y=y, z=z, h=hf, m=mf, rho=rhof, cs=csf, u=u)

    # add requisite properties
    # sound speed etc.
    add_properties(pa, 'e')

    # velocity gradient properties
    add_properties(pa, 'v00', 'v01', 'v02', 'v10', 'v11', 'v12', 'v20', 'v21', 'v22')

    # artificial stress properties
    add_properties(pa, 'r00', 'r01', 'r02', 'r11', 'r12', 'r22')

    # deviatoric stress components
    add_properties(pa, 's00', 's01', 's02', 's11', 's12', 's22')

    # deviatoric stress accelerations
    add_properties(pa, 'as00', 'as01', 'as02', 'as11', 'as12', 'as22')

    # deviatoric stress initial values
    add_properties(pa, 's000', 's010', 's020', 's110', 's120', 's220')

    # standard acceleration variables
    add_properties(pa, 'arho', 'au', 'av', 'aw', 'ax', 'ay', 'az', 'ae')

    # initial values
    add_properties(pa, 'rho0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0', 'e0')

    pa.add_constant('G', G2)
    pa.add_constant('n', 4)

    kernel = Gaussian(dim=3)
    wdeltap = kernel.kernel(rij=dx, h=hdx*dx)
    pa.add_constant('wdeltap', wdeltap)

    # load balancing properties
    pa.set_lb_props( list(pa.properties.keys()) )

    return projectile

def get_plate_particles():
    xarr = numpy.arange(0, 0.002+dx, dx)
    yarr = numpy.arange(-0.020, 0.02+dx, dx)
    zarr = numpy.arange(-0.02, 0.02+dx, dx)

    x, y, z = numpy.mgrid[0:0.002+dx:dx, -0.02:0.02+dx:dx, -0.02:0.02+dx:dx]
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()

    print('%d Target particles'%len(x))

    hf = numpy.ones_like(x) * h
    mf = numpy.ones_like(x) * dx * dy * dz * ro1
    rhof = numpy.ones_like(x) * ro1
    csf = numpy.ones_like(x) * cs1
    pa = plate = get_particle_array(name="plate",
                                    x=x, y=y, z=z, h=hf, m=mf, rho=rhof, cs=csf)

    # add requisite properties
    # sound speed etc.
    add_properties(pa, 'e' )

    # velocity gradient properties
    add_properties(pa, 'v00', 'v01', 'v02', 'v10', 'v11', 'v12', 'v20', 'v21', 'v22')

    # artificial stress properties
    add_properties(pa, 'r00', 'r01', 'r02', 'r11', 'r12', 'r22')

    # deviatoric stress components
    add_properties(pa, 's00', 's01', 's02', 's11', 's12', 's22')

    # deviatoric stress accelerations
    add_properties(pa, 'as00', 'as01', 'as02', 'as11', 'as12', 'as22')

    # deviatoric stress initial values
    add_properties(pa, 's000', 's010', 's020', 's110', 's120', 's220')

    # standard acceleration variables
    add_properties(pa, 'arho', 'au', 'av', 'aw', 'ax', 'ay', 'az', 'ae')

    # initial values
    add_properties(pa, 'rho0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0', 'e0')

    pa.add_constant('G', G1)
    pa.add_constant('n', 4)

    kernel = Gaussian(dim=3)
    wdeltap = kernel.kernel(rij=dx, h=hdx*dx)
    pa.add_constant('wdeltap', wdeltap)
    # load balancing properties
    pa.set_lb_props( list(pa.properties.keys()) )

    # removed S_00 and similar components
    plate.v[:]=0.0
    return plate

class Impact(Application):
    def create_particles(self):
        plate      = get_plate_particles()
        projectile = get_projectile_particles()

        return [plate, projectile]

    def create_solver(self):
        dim=3
        kernel = Gaussian(dim=dim)
        #kernel = WendlandQuintic(dim=dim)

        integrator = EPECIntegrator(projectile=SolidMechStep(), plate=SolidMechStep())
        solver     = Solver(kernel=kernel, dim=dim, integrator=integrator)

        dt = 1e-9
        tf = 8e-6
        solver.set_time_step(dt)
        solver.set_final_time(tf)
        solver.set_print_freq(100)
        return solver

    def create_equations(self):
        equations = [

            # update smoothing length
            # Group(
            #     equations = [
            #         UpdateSmoothingLengthFromVolume(dest='plate',      sources=['plate', 'projectile'], dim=dim, k=hdx),
            #         UpdateSmoothingLengthFromVolume(dest='projectile', sources=['plate', 'projectile'], dim=dim, k=hdx),
            #     ],
            #     update_nnps=True,
            # ),

            # compute properties from the current state
            Group(
                equations = [
                    # EOS (compute the pressure using  one of the EOSs)

                    #MieGruneisenEOS(dest='plate',      sources=None, gamma=gamma1, r0=ro1 , c0=C1, S=S1),
                    #MieGruneisenEOS(dest='projectile', sources=None, gamma=gamma2, r0=ro2 , c0=C2, S=S2),

                    StiffenedGasEOS(dest='plate',      sources=None, gamma=gamma1, r0=ro1 , c0=C1),
                    StiffenedGasEOS(dest='projectile', sources=None, gamma=gamma2, r0=ro2 , c0=C2),

                    # compute the velocity gradient tensor
                    VelocityGradient3D(dest='plate',      sources=['plate']),
                    VelocityGradient3D(dest='projectile', sources=['projectile']),

                    # # stress
                    VonMisesPlasticity2D(dest='plate',      sources=None, flow_stress=Yo1),
                    VonMisesPlasticity2D(dest='projectile', sources=None, flow_stress=Yo2),

                    # # artificial stress to avoid clumping
                    MonaghanArtificialStress(dest='plate',      sources=None, eps=0.3),
                    MonaghanArtificialStress(dest='projectile', sources=None, eps=0.3),

                ]
            ),

            # accelerations (rho, u, v, ...)
            Group(
                equations = [

                    # continuity equation
                    ContinuityEquation(dest='plate',      sources=['projectile','plate']),
                    ContinuityEquation(dest='projectile', sources=['projectile','plate']),

                    # momentum equation
                    MomentumEquationWithStress(dest='projectile', sources=['projectile','plate',]),
                    MomentumEquationWithStress(dest='plate',      sources=['projectile','plate',]),

                    # energy equation:
                    EnergyEquationWithStress(dest='plate',      sources=['projectile','plate',],
                                               alpha=avisc_alpha, beta=avisc_beta, eta=avisc_eta),

                    EnergyEquationWithStress(dest='projectile', sources=['projectile','plate',],
                                               alpha=avisc_alpha, beta=avisc_beta,  eta=avisc_eta),

                    #avisc
                    MonaghanArtificialViscosity(dest='plate',     sources=['projectile','plate'],
                                                alpha=avisc_alpha, beta=avisc_beta),

                    MonaghanArtificialViscosity(dest='projectile', sources=['projectile','plate'],
                                               alpha=avisc_alpha, beta=avisc_beta),

                    #updates to the stress term
                    HookesDeviatoricStressRate(dest='plate',      sources=None),
                    HookesDeviatoricStressRate(dest='projectile', sources=None),

                    #position stepping
                    XSPHCorrection(dest='plate',      sources=['plate'],      eps=xsph_eps),
                    XSPHCorrection(dest='projectile', sources=['projectile'], eps=xsph_eps),

                ]
            ),

        ] # End Group list

        return equations


if __name__ == '__main__':
    app = Impact()
    app.run()



"""
Tests for the particle array module.
"""

# standard imports
import unittest
import numpy

# local imports
from pysph.base import particle_array
from pysph.base import utils

from cyarray.carray import LongArray, IntArray, DoubleArray

import pickle
import pytest

from compyle.config import get_config


def check_array(x, y):
    """Check if two arrays are equal with an absolute tolerance of
    1e-16."""
    return numpy.allclose(x, y, atol=1e-16, rtol=0)


###############################################################################
# `ParticleArrayTest` class.
###############################################################################
class ParticleArrayTest(object):
    """
    Tests for the particle array class.
    """

    def test_constructor(self):
        # Default constructor test.
        p = particle_array.ParticleArray(name='test_particle_array',
                                         backend=self.backend)

        self.assertEqual(p.name, 'test_particle_array')
        self.assertEqual('tag' in p.properties, True)
        self.assertEqual(p.properties['tag'].length, 0)

        # Constructor with some properties.
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        p = particle_array.ParticleArray(x={'data': x},
                                         y={'data': y},
                                         z={'data': z},
                                         m={'data': m},
                                         h={'data': h},
                                         backend=self.backend)

        self.assertEqual(p.name, '')

        self.assertEqual('x' in p.properties, True)
        self.assertEqual('y' in p.properties, True)
        self.assertEqual('z' in p.properties, True)
        self.assertEqual('m' in p.properties, True)
        self.assertEqual('h' in p.properties, True)

        # get the properties are check if they are the same
        xarr = p.properties['x'].get_npy_array()
        self.assertEqual(check_array(xarr, x), True)

        yarr = p.properties['y'].get_npy_array()
        self.assertEqual(check_array(yarr, y), True)

        zarr = p.properties['z'].get_npy_array()
        self.assertEqual(check_array(zarr, z), True)

        marr = p.properties['m'].get_npy_array()
        self.assertEqual(check_array(marr, m), True)

        harr = p.properties['h'].get_npy_array()
        self.assertEqual(check_array(harr, h), True)

        # check if the 'tag' array was added.
        self.assertEqual('tag' in p.properties, True)
        self.assertEqual(list(p.properties.values())[0].length == len(x), True)

        # Constructor with tags
        tags = [0, 1, 0, 1]
        p = particle_array.ParticleArray(x={'data': x}, y={'data': y},
                                         z={'data': z},
                                         tag={'data': tags, 'type': 'int'},
                                         backend=self.backend)
        self.assertEqual(check_array(p.get('tag', only_real_particles=False),
                                     [0, 0, 1, 1]), True)
        self.assertEqual(check_array(p.get('x', only_real_particles=False),
                                     [1, 3, 2, 4]), True)
        self.assertEqual(check_array(p.get('y', only_real_particles=False),
                                     [0, 2, 1, 3]), True)
        self.assertEqual(check_array(p.get('z', only_real_particles=False),
                                     [0, 0, 0, 0]), True)

        # trying to create particle array without any values but some
        # properties.
        p = particle_array.ParticleArray(x={}, y={}, z={}, h={},
                                         backend=self.backend)

        self.assertEqual(p.get_number_of_particles(), 0)
        self.assertEqual('x' in p.properties, True)
        self.assertEqual('y' in p.properties, True)
        self.assertEqual('z' in p.properties, True)
        self.assertEqual('tag' in p.properties, True)

        # now trying to supply some properties with values and others without
        p = particle_array.ParticleArray(
            x={'default': 10.0}, y={'data': [1.0, 2.0]},
            z={}, h={'data': [0.1, 0.1]}, backend=self.backend
        )

        self.assertEqual(p.get_number_of_particles(), 2)
        self.assertEqual(check_array(p.x, [10., 10.]), True)
        self.assertEqual(check_array(p.y, [1., 2.]), True)
        self.assertEqual(check_array(p.z, [0, 0]), True)
        self.assertEqual(check_array(p.h, [0.1, 0.1]), True)

    def test_constructor_works_with_strides(self):
        # Given
        x = [1, 2, 3, 4.]
        rho = 10.0
        data = numpy.arange(8)

        # When
        p = particle_array.ParticleArray(
            x=x, rho=rho, data={'data': data, 'stride': 2}, name='fluid',
            backend=self.backend
        )

        # Then
        self.assertEqual(p.name, 'fluid')

        self.assertEqual('x' in p.properties, True)
        self.assertEqual('rho' in p.properties, True)
        self.assertEqual('data' in p.properties, True)
        self.assertEqual('tag' in p.properties, True)
        self.assertEqual('pid' in p.properties, True)
        self.assertEqual('gid' in p.properties, True)

        # get the properties are check if they are the same
        self.assertEqual(check_array(p.x, x), True)
        self.assertEqual(check_array(p.rho, numpy.ones(4) * rho), True)
        self.assertEqual(check_array(p.data, numpy.ravel(data)), True)

    def test_constructor_works_with_simple_props(self):
        # Given
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        rho = 10.0
        data = numpy.diag((2, 2))

        # When
        p = particle_array.ParticleArray(
            x=x, y=y, rho=rho, data=data, name='fluid',
            backend=self.backend
        )

        # Then
        self.assertEqual(p.name, 'fluid')

        self.assertEqual('x' in p.properties, True)
        self.assertEqual('y' in p.properties, True)
        self.assertEqual('rho' in p.properties, True)
        self.assertEqual('data' in p.properties, True)
        self.assertEqual('tag' in p.properties, True)
        self.assertEqual('pid' in p.properties, True)
        self.assertEqual('gid' in p.properties, True)

        # get the properties are check if they are the same
        self.assertEqual(check_array(p.x, x), True)
        self.assertEqual(check_array(p.y, y), True)
        self.assertEqual(check_array(p.rho, numpy.ones(4) * rho), True)
        self.assertEqual(check_array(p.data, numpy.ravel(data)), True)

    def test_get_number_of_particles(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, A={'data': A, 'stride': 3},
            backend=self.backend
        )

        self.assertEqual(p.get_number_of_particles(), 4)

    def test_get(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, A={'data': A, 'stride': 3},
            backend=self.backend
        )

        self.assertEqual(check_array(x, p.get('x')), True)
        self.assertEqual(check_array(y, p.get('y')), True)
        self.assertEqual(check_array(z, p.get('z')), True)
        self.assertEqual(check_array(m, p.get('m')), True)
        self.assertEqual(check_array(h, p.get('h')), True)
        self.assertEqual(check_array(A.ravel(), p.get('A')), True)

    def test_clear(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]

        p = particle_array.ParticleArray(x={'data': x}, y={'data': y},
                                         z={'data': z}, m={'data': m},
                                         h={'data': h}, backend=self.backend)

        p.clear()

        self.assertEqual(len(p.properties), 3)
        self.assertEqual('tag' in p.properties, True)
        self.assertEqual(p.properties['tag'].length, 0)

        self.assertEqual('pid' in p.properties, True)
        self.assertEqual(p.properties['pid'].length, 0)

        self.assertEqual('gid' in p.properties, True)
        self.assertEqual(p.properties['gid'].length, 0)

    def test_getattr(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, A={'data': A, 'stride': 3},
            backend=self.backend
        )

        self.assertEqual(check_array(x, p.x), True)
        self.assertEqual(check_array(y, p.y), True)
        self.assertEqual(check_array(z, p.z), True)
        self.assertEqual(check_array(m, p.m), True)
        self.assertEqual(check_array(h, p.h), True)
        self.assertEqual(check_array(A.ravel(), p.get('A')), True)

        # try getting an non-existant attribute
        self.assertRaises(AttributeError, p.__getattr__, 'a')

    def test_setattr(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, A={'data': A, 'stride': 3},
            backend=self.backend
        )

        p.x = p.x * 2.0

        self.assertEqual(check_array(p.get('x'), [2., 4, 6, 8]), True)
        p.x = p.x + 3.0 * p.x
        self.assertEqual(check_array(p.get('x'), [8., 16., 24., 32.]), True)
        p.A = p.A*2
        self.assertEqual(check_array(p.get('A').ravel(), A*2), True)

    def test_remove_particles(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, A={'data': A, 'stride': 3},
            backend=self.backend
        )

        remove_arr = LongArray(0)
        remove_arr.append(0)
        remove_arr.append(1)

        p.remove_particles(remove_arr)
        self.pull(p)

        self.assertEqual(p.get_number_of_particles(), 2)
        self.assertEqual(check_array(p.x, [3., 4.]), True)
        self.assertEqual(check_array(p.y, [2., 3.]), True)
        self.assertEqual(check_array(p.z, [0., 0.]), True)
        self.assertEqual(check_array(p.m, [1., 1.]), True)
        self.assertEqual(check_array(p.h, [.1, .1]), True)
        self.assertEqual(check_array(p.A, numpy.arange(6, 12)), True)

        # now try invalid operations to make sure errors are raised.
        remove_arr.resize(10)
        self.assertRaises(ValueError, p.remove_particles, remove_arr)

        # now try to remove a particle with index more that particle
        # length.
        remove_arr = [2]

        p.remove_particles(remove_arr)
        self.pull(p)
        # make sure no change occurred.
        self.assertEqual(p.get_number_of_particles(), 2)
        self.assertEqual(check_array(p.x, [3., 4.]), True)
        self.assertEqual(check_array(p.y, [2., 3.]), True)
        self.assertEqual(check_array(p.z, [0., 0.]), True)
        self.assertEqual(check_array(p.m, [1., 1.]), True)
        self.assertEqual(check_array(p.h, [.1, .1]), True)
        self.assertEqual(check_array(p.A, numpy.arange(6, 12)), True)

    def test_add_particles(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, A=dict(data=A, stride=3),
            backend=self.backend
        )

        new_particles = {}
        new_particles['x'] = numpy.array([5., 6, 7], dtype=numpy.float32)
        new_particles['y'] = numpy.array([4., 5, 6], dtype=numpy.float32)
        new_particles['z'] = numpy.array([0., 0, 0], dtype=numpy.float32)

        p.add_particles(**new_particles)
        self.pull(p)

        self.assertEqual(p.get_number_of_particles(), 7)
        self.assertEqual(check_array(p.x, [1., 2, 3, 4, 5, 6, 7]), True)
        self.assertEqual(check_array(p.y, [0., 1, 2, 3, 4, 5, 6]), True)
        self.assertEqual(check_array(p.z, [0., 0, 0, 0, 0, 0, 0]), True)
        expect = numpy.zeros(21, dtype=A.dtype)
        expect[:12] = A
        numpy.testing.assert_array_equal(p.A, expect)

        # make sure the other arrays were resized
        self.assertEqual(len(p.h), 7)
        self.assertEqual(len(p.m), 7)

        # try adding an empty particle list
        p.add_particles(**{})
        self.pull(p)
        self.assertEqual(p.get_number_of_particles(), 7)
        self.assertEqual(check_array(p.x, [1., 2, 3, 4, 5, 6, 7]), True)
        self.assertEqual(check_array(p.y, [0., 1, 2, 3, 4, 5, 6]), True)
        self.assertEqual(check_array(p.z, [0., 0, 0, 0, 0, 0, 0]), True)
        self.assertEqual(check_array(p.A, expect), True)

        # make sure the other arrays were resized
        self.assertEqual(len(p.h), 7)
        self.assertEqual(len(p.m), 7)

        # adding particles with tags
        p = particle_array.ParticleArray(x={'data': x}, y={'data': y},
                                         z={'data': z}, m={'data': m},
                                         h={'data': h},
                                         backend=self.backend)
        p.add_particles(x=[5, 6, 7, 8], tag=[1, 1, 0, 0])
        self.pull(p)

        self.assertEqual(p.get_number_of_particles(), 8)
        self.assertEqual(check_array(p.x, [1, 2, 3, 4, 7, 8]), True)
        self.assertEqual(check_array(p.y, [0, 1, 2, 3, 0, 0]), True)
        self.assertEqual(check_array(p.z, [0, 0, 0, 0, 0, 0]), True)

    def test_remove_tagged_particles(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        A = numpy.arange(12)
        tag = [1, 1, 1, 0]

        p = particle_array.ParticleArray(
            x={'data': x}, y={'data': y},
            z={'data': z}, m={'data': m},
            h={'data': h}, tag={'data': tag}, A={'data': A, 'stride': 3},
            backend=self.backend
        )

        numpy.testing.assert_array_equal(
            p.get('A'), numpy.arange(9, 12)
        )

        p.remove_tagged_particles(0)
        self.pull(p)

        self.assertEqual(p.get_number_of_particles(), 3)
        self.assertEqual(
            check_array(
                numpy.sort(p.get('x', only_real_particles=False)),
                [1., 2., 3.]),
            True
        )
        self.assertEqual(
            check_array(
                numpy.sort(p.get('y', only_real_particles=False)),
                [0., 1., 2.]
            ), True
        )
        self.assertEqual(
            check_array(
                numpy.sort(p.get('z', only_real_particles=False)),
                [0., 0, 0]
            ), True
        )
        self.assertEqual(
            check_array(
                numpy.sort(p.get('h', only_real_particles=False)),
                [0.1, 0.1, 0.1]
            ), True
        )
        self.assertEqual(
            check_array(p.get('m', only_real_particles=False), [1., 1., 1.]),
            True
        )
        if p.gpu is None:
            numpy.testing.assert_array_equal(
                p.get('A', only_real_particles=False),
                numpy.arange(9)
            )
        else:
            numpy.testing.assert_array_equal(
                p.get('A', only_real_particles=False),
                list(range(3, 9)) + [0., 1, 2]
            )
        self.assertEqual(check_array(p.x, []), True)
        self.assertEqual(check_array(p.y, []), True)
        self.assertEqual(check_array(p.z, []), True)
        self.assertEqual(check_array(p.h, []), True)
        self.assertEqual(check_array(p.m, []), True)
        self.assertEqual(check_array(p.A, []), True)

    def test_add_property(self):
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]
        tag = [0, 0, 0, 0]

        p = particle_array.ParticleArray(x={'data': x}, y={'data': y},
                                         z={'data': z}, m={'data': m},
                                         h={'data': h}, tag={'data': tag},
                                         backend=self.backend)

        p.add_property(**{'name': 'x'})
        # make sure the current 'x' property is intact.
        self.assertEqual(check_array(p.x, x), True)

        # add a property with complete specification
        p.add_property(**{'name': 'f1',
                          'data': [1, 1, 2, 3],
                          'type': 'int',
                          'default': 4})
        self.assertEqual(check_array(p.f1, [1, 1, 2, 3]), True)
        self.assertEqual(type(p.properties['f1']), IntArray)
        self.assertEqual(p.default_values['f1'], 4)

        # add a property with stride.
        data = [1, 1, 2, 2, 3, 3, 4, 4]
        p.add_property(**{'name': 'm1',
                          'data': data,
                          'type': 'int',
                          'stride': 2})
        self.assertEqual(check_array(p.m1, data), True)
        self.assertEqual(type(p.properties['m1']), IntArray)

        # add a property without specifying the type
        p.add_property(**{'name': 'f2',
                          'data': [1, 1, 2, 3],
                          'default': 4.0})
        self.assertEqual(type(p.properties['f2']), DoubleArray)
        self.assertEqual(check_array(p.f2, [1, 1, 2, 3]), True)

        p.add_property(**{'name': 'f3'})
        self.assertEqual(type(p.properties['f3']), DoubleArray)
        self.assertEqual(p.properties['f3'].length,
                         p.get_number_of_particles())
        self.assertEqual(check_array(p.f3, [0, 0, 0, 0]), True)

        p.add_property(**{'name': 'f4', 'default': 3.0})
        self.assertEqual(type(p.properties['f4']), DoubleArray)
        self.assertEqual(p.properties['f4'].length,
                         p.get_number_of_particles())
        self.assertEqual(check_array(p.f4, [3, 3, 3, 3]), True)

        p.add_property('f5', data=10.0)
        self.assertEqual(type(p.properties['f5']), DoubleArray)
        self.assertEqual(p.properties['f5'].length,
                         p.get_number_of_particles())
        self.assertEqual(check_array(p.f5, [10.0, 10.0, 10.0, 10.0]), True)

        p.add_property('m2', data=10.0, stride=2)
        self.assertEqual(type(p.properties['m2']), DoubleArray)
        self.assertEqual(p.properties['m2'].length,
                         p.get_number_of_particles()*2)
        self.assertEqual(check_array(p.m2, [10.0]*8), True)

    def test_extend(self):
        # Given
        p = particle_array.ParticleArray(default_particle_tag=10, x={},
                                         y={'default': -1.},
                                         backend=self.backend)
        p.add_property('A', default=5.0, stride=2)

        # When
        p.extend(5)
        p.align_particles()
        self.pull(p)

        # Then
        self.assertEqual(p.get_number_of_particles(), 5)
        self.assertEqual(check_array(p.get(
            'x', only_real_particles=False), [0, 0, 0, 0, 0]), True)
        self.assertEqual(check_array(p.get('y', only_real_particles=False),
                                     [-1., -1., -1., -1., -1.]), True)
        self.assertEqual(check_array(p.get('tag', only_real_particles=False),
                                     [10, 10, 10, 10, 10]), True)
        self.assertEqual(check_array(p.get('A', only_real_particles=False),
                                     [5.0]*10), True)

        # Given
        p = particle_array.ParticleArray(
            A={'data': [10.0, 10.0], 'stride': 2, 'default': -1.},
            backend=self.backend
        )

        # When
        p.extend(5)
        p.align_particles()
        self.pull(p)

        # Then
        self.assertEqual(check_array(p.get('A', only_real_particles=False),
                                     [10.0, 10.0] + [-1.0]*10), True)

    def test_resize(self):
        # Given
        p = particle_array.ParticleArray(
            A={'data': [10.0, 10.0], 'stride': 2, 'default': -1.},
            x=[1.0], backend=self.backend
        )

        # When
        p.resize(4)
        p.align_particles()
        self.pull(p)

        # Then
        self.assertEqual(p.get_carray('x').length, 4)
        self.assertEqual(p.get_carray('A').length, 8)

    def test_align_particles(self):
        # Given
        p = particle_array.ParticleArray(backend=self.backend)
        p.add_property(**{'name': 'x',
                          'data': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        p.add_property(**{'name': 'y',
                          'data': [10, 9, 8, 7, 6, 5, 4, 3, 2, 1]})
        a = numpy.arange(10) + 1
        A = numpy.zeros(20)
        A[::2] = a
        A[1::2] = a
        p.add_property('A', data=A, stride=2)
        print(A)
        p.set(**{'tag': [0, 0, 1, 1, 1, 0, 4, 0, 1, 5]})

        # When
        self.push(p)
        p.align_particles()
        self.pull(p)

        # Then
        x_new = p.get('x', only_real_particles=False)
        y_new = p.get('y', only_real_particles=False)
        A_new = p.get('A', only_real_particles=False)
        print(A_new)

        # check the local particles
        self.assertEqual(check_array(x_new[:4], [1, 2, 6, 8]),
                         True)
        self.assertEqual(check_array(y_new[:4], [10, 9, 5, 3]),
                         True)
        self.assertEqual(check_array(A_new[:8:2], [1, 2, 6, 8]),
                         True)
        self.assertEqual(check_array(A_new[1:8:2], [1, 2, 6, 8]),
                         True)

        # check the remaining particles
        self.assertEqual(
            check_array(numpy.sort(x_new[4:]),
                        [3, 4, 5, 7, 9, 10]), True
        )
        self.assertEqual(
            check_array(numpy.sort(y_new[4:]),
                        [1, 2, 4, 6, 7, 8]), True
        )
        self.assertEqual(
            check_array(numpy.sort(A_new[8::2]),
                        [3, 4, 5, 7, 9, 10]), True
        )
        self.assertEqual(
            check_array(numpy.sort(A_new[9::2]),
                        [3, 4, 5, 7, 9, 10]), True
        )

        p.set(**{'tag': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1]})
        self.push(p)
        p.align_particles()
        self.pull(p)
        x_new = p.get('x', only_real_particles=False)
        y_new = p.get('y', only_real_particles=False)

        # check the remaining particles
        self.assertEqual(check_array(x_new[:4], [1, 2, 6, 8]),
                         True)
        self.assertEqual(check_array(y_new[:4], [10, 9, 5, 3]),
                         True)
        self.assertEqual(check_array(A_new[:8:2], [1, 2, 6, 8]),
                         True)
        self.assertEqual(check_array(A_new[1:8:2], [1, 2, 6, 8]),
                         True)

        # check the remaining particles
        self.assertEqual(
            check_array(numpy.sort(x_new[4:]),
                        [3, 4, 5, 7, 9, 10]), True
        )
        self.assertEqual(
            check_array(numpy.sort(y_new[4:]),
                        [1, 2, 4, 6, 7, 8]), True
        )
        self.assertEqual(
            check_array(numpy.sort(A_new[8::2]),
                        [3, 4, 5, 7, 9, 10]), True
        )
        self.assertEqual(
            check_array(numpy.sort(A_new[9::2]),
                        [3, 4, 5, 7, 9, 10]), True
        )

    def test_append_parray(self):
        # Given
        p1 = particle_array.ParticleArray(backend=self.backend)
        p1.add_property(**{'name': 'x', 'data': [1, 2, 3]})
        p1.add_property('A', data=2.0, stride=2)
        p2 = particle_array.ParticleArray(x={'data': [4, 5, 6]},
                                          y={'data': [1, 2, 3]},
                                          tag={'data': [1, 0, 1]},
                                          backend=self.backend)

        # When
        p1.append_parray(p2)
        self.pull(p1)

        # Then
        self.assertEqual(p1.get_number_of_particles(), 6)
        self.assertEqual(check_array(p1.x, [1, 2, 3, 5]), True)
        self.assertEqual(check_array(p1.y, [0, 0, 0, 2]), True)
        numpy.testing.assert_array_equal(p1.A, [2.0]*6 + [0.0]*2)
        self.assertEqual(check_array(p1.tag, [0, 0, 0, 0]), True)

        # Given
        p1 = particle_array.ParticleArray(backend=self.backend)
        p1.add_property(**{'name': 'x', 'data': [1, 2, 3]})
        # In this case the new strided prop is in the second parray.
        p2 = particle_array.ParticleArray(x={'data': [4, 5, 6]},
                                          y={'data': [1, 2, 3]},
                                          tag={'data': [1, 0, 1]},
                                          backend=self.backend)
        p2.add_property('A', data=2.0, stride=2)

        # When
        p1.append_parray(p2)
        self.pull(p1)

        # Then
        self.assertEqual(p1.get_number_of_particles(), 6)
        self.assertEqual(check_array(p1.x, [1, 2, 3, 5]), True)
        self.assertEqual(check_array(p1.y, [0, 0, 0, 2]), True)
        self.assertEqual(check_array(p1.A, [0.0]*6 + [2.0]*2), True)
        self.assertEqual(check_array(p1.tag, [0, 0, 0, 0]), True)

    def test_copy_properties(self):
        # Given
        p1 = particle_array.ParticleArray(backend=self.backend)
        p1.add_property(**{'name': 'x',
                           'data': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})
        p1.add_property(name='y')
        p1.add_property(name='t')
        p1.add_property('A', data=2.0, stride=2)

        p2 = particle_array.ParticleArray(backend=self.backend)
        p2.add_property(name='t', data=[-1, -1, -1, -1])
        p2.add_property(name='s', data=[2, 3, 4, 5])
        p2.add_property('A', data=3.0, stride=2)

        # When
        p1.copy_properties(p2, start_index=5, end_index=9)

        # Then
        self.assertEqual(check_array(p1.t, [0, 0, 0, 0, 0, -1, -1, -1, -1, 0]),
                         True)
        numpy.testing.assert_array_equal(
            p1.A, [2.0]*10 + [3.0]*8 + [2.0]*2
        )

        # When
        p1.add_property('s')
        p1.copy_properties(p2, start_index=5, end_index=9)

        # Then
        self.assertEqual(check_array(p1.t, [0, 0, 0, 0, 0, -1, -1, -1, -1, 0]),
                         True)
        self.assertEqual(
            check_array(p1.s, [0, 0, 0, 0, 0, 2, 3, 4, 5, 0]), True
        )
        numpy.testing.assert_array_equal(
            p1.A, [2.0]*10 + [3.0]*8 + [2.0]*2
        )

    def test_that_constants_can_be_added(self):
        # Given
        p = particle_array.ParticleArray(backend=self.backend)
        nprop = len(p.properties)
        self.assertEqual(len(p.constants), 0)

        # When
        p.add_constant('s', 0.0)
        p.add_constant('ii', 0)
        p.add_constant('v', [0.0, 1.0, 2.0])

        # Then
        self.assertEqual(len(p.constants), 3)
        self.assertEqual(len(p.properties), nprop)
        self.assertEqual(p.s[0], 0.0)
        self.assertEqual(p.ii[0], 0)
        self.assertTrue(str(p.ii[0].dtype).startswith('int'))
        self.assertTrue(check_array(p.v, [0.0, 1.0, 2.0]))

    def test_that_constants_can_be_set_in_constructor(self):
        # Given
        # When
        p = particle_array.ParticleArray(
            constants=dict(s=0.0, v=[0.0, 1.0, 2.0]),
            backend=self.backend
        )
        nprop = len(p.properties)

        # Then
        self.assertEqual(len(p.constants), 2)
        self.assertEqual(len(p.properties), nprop)
        self.assertEqual(p.s[0], 0.0)
        self.assertTrue(check_array(p.v, [0.0, 1.0, 2.0]))

    def test_that_get_works_on_constants(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3],
                                         backend=self.backend)

        # When
        p.add_constant('s', 0.0)
        p.add_constant('v', [0.0, 1.0, 2.0])

        # Then
        self.assertTrue(check_array(p.get('s'), [0.0]))
        self.assertTrue(check_array(p.get('v'), [0.0, 1.0, 2.0]))

    def test_that_constants_are_not_resized_when_particles_are_added(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1.0],
                                         backend=self.backend)

        # When
        p.add_constant('v', [0.0, 1.0])
        p.add_particles(x=[2.0, 3.0])
        self.pull(p)

        # Then
        self.assertTrue(check_array(p.v, [0.0, 1.0]))
        self.assertTrue(check_array(p.x, [1.0, 2.0, 3.0]))

    def test_that_set_works_on_constants(self):
        # Given
        constants = dict(v=[0.0, 0.0, 0.0], c=[0.0, 0.0, 0.0])
        p = particle_array.ParticleArray(name='f', constants=constants,
                                         backend=self.backend)

        # When
        p.set(v=[0.0, 1.0, 2.0])
        p.c = [0.0, 1.0, 2.0]

        # Then
        self.assertEqual(len(p.constants), 2)
        self.assertTrue(check_array(p.get('v'), [0.0, 1.0, 2.0]))
        self.assertTrue(check_array(p.get('c'), [0.0, 1.0, 2.0]))

    def test_that_get_carray_works_with_constants(self):
        # Given
        p = particle_array.ParticleArray(backend=self.backend)
        v = [0.0, 1.0, 2.0]

        # When
        p.add_constant('v', v)
        a = p.get_carray('v')

        # Then
        self.assertEqual(a.get_c_type(), 'double')
        self.assertTrue(check_array(a.get_npy_array(), v))

    def test_empty_clone_works_without_specific_props(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)
        p.set_output_arrays(['x', 'A'])
        v = [0.0, 1.0, 2.0]
        p.add_constant('v', v)

        # When
        clone = p.empty_clone()
        self.pull(clone)

        # Then
        self.assertTrue(check_array(clone.v, v))
        self.assertEqual(sorted(clone.output_property_arrays),
                         sorted(p.output_property_arrays))

    def test_empty_clone_works_with_specific_props(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)
        p.set_output_arrays(['x', 'A'])
        v = [0.0, 1.0, 2.0]
        p.add_constant('v', v)

        # When
        clone = p.empty_clone(props=['x', 'A'])
        self.pull(clone)

        # Then
        self.assertTrue(check_array(clone.v, v))
        self.assertEqual(sorted(clone.output_property_arrays),
                         sorted(p.output_property_arrays))
        self.assertFalse('y' in clone.properties)
        self.assertTrue('x' in clone.properties)
        self.assertTrue('A' in clone.properties)

    def test_extract_particles_works_without_specific_props_without_dest(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)

        # When.
        n = p.extract_particles([1])
        self.pull(n)

        # Then.
        self.assertEqual(len(p.x), 3)
        self.assertEqual(len(p.A), 6)
        self.assertEqual(len(n.x), 1)
        self.assertEqual(len(n.A), 2)
        self.assertEqual(n.x[0], 2.0)
        numpy.testing.assert_array_equal(n.A, [2, 3])

    def test_extract_particles_works_with_specific_props_without_dest(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3], y=[0, 0, 0],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)

        # When.
        n = p.extract_particles([1, 2], props=['x', 'A'])
        self.pull(n)

        # Then.
        self.assertEqual(len(p.x), 3)
        self.assertEqual(len(n.x), 2)
        self.assertEqual(n.x[0], 2.0)
        self.assertEqual(n.x[0], 2.0)
        numpy.testing.assert_array_equal(n.A, [2, 3, 4, 5])
        self.assertFalse('y' in n.properties)

    def test_extract_particles_works_without_specific_props_with_dest(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)
        n = p.empty_clone()

        # When.
        p.extract_particles([1], dest_array=n)
        self.pull(n)

        # Then.
        self.assertEqual(len(p.x), 3)
        self.assertEqual(len(p.A), 6)
        self.assertEqual(len(n.x), 1)
        self.assertEqual(len(n.A), 2)
        self.assertEqual(n.x[0], 2.0)
        numpy.testing.assert_array_equal(n.A, [2, 3])

    def test_extract_particles_works_with_non_empty_dest(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)
        n = p.empty_clone()
        n.extend(2)

        # When.
        p.extract_particles([1], dest_array=n)
        self.pull(n)

        # Then.
        self.assertEqual(len(p.x), 3)
        self.assertEqual(len(p.A), 6)
        self.assertEqual(len(n.x), 3)
        self.assertEqual(len(n.A), 6)
        numpy.testing.assert_array_equal(n.x, [0.0, 0.0, 2.0])
        numpy.testing.assert_array_equal(n.A, [0, 0, 0, 0, 2, 3])

    def test_extract_particles_works_with_specific_props_with_dest(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3], y=[0, 0, 0],
                                         backend=self.backend)
        p.add_property('A', data=numpy.arange(6), stride=2)
        n = p.empty_clone(props=['x', 'A'])

        # When.
        p.extract_particles([1, 2], dest_array=n, props=['x', 'A'])
        self.pull(n)

        # Then.
        self.assertEqual(len(p.x), 3)
        self.assertEqual(len(n.x), 2)
        self.assertEqual(n.x[0], 2.0)
        self.assertEqual(n.x[0], 2.0)
        numpy.testing.assert_array_equal(n.A, [2, 3, 4, 5])
        self.assertFalse('y' in n.properties)

    def test_that_remove_property_also_removes_output_arrays(self):
        # Given
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3], y=[0, 0, 0],
                                         backend=self.backend)
        p.add_property('test')
        p.set_output_arrays(['x', 'y', 'test'])

        # When
        p.remove_property('test')

        # Then
        self.assertEqual(p.output_property_arrays, ['x', 'y'])


class ParticleArrayUtils(unittest.TestCase):
    def setUp(self):
        get_config().use_opencl = False

    def test_that_get_particles_info_works(self):
        # Given.
        p = particle_array.ParticleArray(name='f', x=[1, 2, 3])
        p.add_property('A', data=numpy.arange(6), stride=2)
        c = [1.0, 2.0]
        p.add_constant('c', c)
        p.set_lb_props(['x'])

        # When.
        info = utils.get_particles_info([p])
        pas = utils.create_dummy_particles(info)
        dummy = pas[0]

        # Then.
        self.assertTrue(check_array(dummy.c, c))
        self.assertEqual(dummy.name, 'f')
        self.assertTrue('x' in dummy.properties)
        self.assertTrue('A' in dummy.properties)
        self.assertTrue('A' in dummy.stride)
        self.assertEqual(dummy.stride['A'], 2)
        self.assertEqual(dummy.get_lb_props(), p.get_lb_props())

    def test_get_particle_array_takes_scalars(self):
        # Given/when
        x = [1, 2, 3, 4]
        data = numpy.diag((2, 2))
        pa = utils.get_particle_array(x=x, y=0, rho=1, data=data)

        # Then
        self.assertTrue(numpy.allclose(x, pa.x))
        self.assertTrue(numpy.allclose(numpy.zeros(4), pa.y))
        self.assertTrue(numpy.allclose(numpy.ones(4), pa.rho))
        self.assertTrue(numpy.allclose(numpy.ravel(data), pa.data))


class ParticleArrayTestCPU(unittest.TestCase, ParticleArrayTest):
    """
    Tests for the particle array class.
    """

    def setUp(self):
        get_config().use_opencl = False
        self.backend = None

    def pull(self, p):
        pass

    def push(self, p):
        pass

    def test_pickle(self):
        """
        Tests the pickle and unpickle functions
        """
        p1 = particle_array.ParticleArray()
        p1.add_property('x', data=numpy.arange(10))
        p1.add_property('y', data=numpy.arange(10))
        p1.add_constant('c', [0.0, 1.0])
        p1.align_particles()

        s = pickle.dumps(p1)
        p2 = pickle.loads(s)

        self.assertEqual(len(p1.x), len(p2.x))
        check_array(p1.x, p2.x)
        self.assertEqual(len(p1.y), len(p2.y))
        check_array(p1.y, p2.y)
        self.assertEqual(len(p1.c), len(p2.c))
        check_array(p1.c, p2.c)

    def test_set(self):
        """
        Tests the set function.
        """
        x = [1, 2, 3, 4.]
        y = [0., 1., 2., 3.]
        z = [0., 0., 0., 0.]
        m = [1., 1., 1., 1.]
        h = [.1, .1, .1, .1]

        p = particle_array.ParticleArray(x={'data': x}, y={'data': y},
                                         z={'data': z}, m={'data': m},
                                         h={'data': h},
                                         backend=self.backend)

        # set the x array with new values
        p.set(**{'x': [4., 3, 2, 1], 'h': [0.2, 0.2, 0.2, 0.2]})
        self.assertEqual(check_array(p.get('x'), [4., 3, 2, 1]), True)
        self.assertEqual(check_array(p.get('h'), [0.2, 0.2, 0.2, 0.2]), True)

        # trying to set the tags
        p.set(**{'tag': [0, 1, 1, 1]})
        p.align_particles()
        self.pull(p)
        self.assertEqual(
            check_array(p.get('tag', only_real_particles=False), [0, 1, 1, 1]),
            True
        )
        self.assertEqual(check_array(p.get('tag'), [0]), True)

        # try setting array with smaller length array.
        p.set(**{'x': [5, 6, 7]})
        self.assertEqual(check_array(p.get('x', only_real_particles=False),
                                     [5, 6, 7, 1]), True)
        # try setting array with longer array.
        self.assertRaises(ValueError, p.set, **{'x': [1., 2, 3, 5, 6]})


class ParticleArrayTestOpenCL(unittest.TestCase, ParticleArrayTest):
    def setUp(self):
        ocl = pytest.importorskip("pyopencl")
        cfg = get_config()
        self.orig_use_double = cfg.use_double
        cfg.use_double = True
        self.backend = 'opencl'

    def tearDown(self):
        get_config().use_double = self.orig_use_double

    def pull(self, p):
        p.gpu.pull()

    def push(self, p):
        p.gpu.push()


class ParticleArrayTestCUDA(unittest.TestCase, ParticleArrayTest):
    def setUp(self):
        cu = pytest.importorskip("pycuda")
        cfg = get_config()
        self.orig_use_double = cfg.use_double
        cfg.use_double = True
        self.backend = 'cuda'

    def tearDown(self):
        get_config().use_double = self.orig_use_double

    def pull(self, p):
        p.gpu.pull()

    def push(self, p):
        p.gpu.push()


if __name__ == '__main__':
    import logging
    logger = logging.getLogger()
    ch = logging.StreamHandler()
    logger.addHandler(ch)

    unittest.main()



"""Evolution of a circular patch of incompressible fluid. (60 seconds)

This shows how one can explicitly setup equations and the solver instead of
using a scheme.
"""
from __future__ import print_function
from numpy import ones_like, mgrid

# PySPH base and carray imports
from pysph.base.utils import get_particle_array_wcsph
from pysph.base.kernels import Gaussian

from pysph.solver.solver import Solver
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.integrator_step import WCSPHStep

from pysph.sph.equation import Group
from pysph.sph.basic_equations import XSPHCorrection, ContinuityEquation
from pysph.sph.wc.basic import TaitEOS, MomentumEquation

from pysph.examples.elliptical_drop import EllipticalDrop as EDScheme


class EllipticalDrop(EDScheme):
    def create_scheme(self):
        # Don't create a scheme as done in the parent example class.
        return None

    def create_particles(self):
        """Create the circular patch of fluid."""
        dx = self.dx
        hdx = self.hdx
        ro = self.ro
        name = 'fluid'
        x, y = mgrid[-1.05:1.05+1e-4:dx, -1.05:1.05+1e-4:dx]
        # Get the particles inside the circle.
        condition = ~((x*x + y*y - 1.0) > 1e-10)
        x = x[condition].ravel()
        y = y[condition].ravel()

        m = ones_like(x)*dx*dx*ro
        h = ones_like(x)*hdx*dx
        rho = ones_like(x) * ro
        u = -100*x
        v = 100*y

        pa = get_particle_array_wcsph(x=x, y=y, m=m, rho=rho, h=h,
                                      u=u, v=v, name=name)

        print("Elliptical drop :: %d particles" %
              (pa.get_number_of_particles()))

        # add requisite variables needed for this formulation
        for name in ('arho', 'au', 'av', 'aw', 'ax', 'ay', 'az', 'rho0', 'u0',
                     'v0', 'w0', 'x0', 'y0', 'z0'):
            pa.add_property(name)

        # set the output property arrays
        pa.set_output_arrays(['x', 'y', 'u', 'v', 'rho', 'm',
                              'h', 'p', 'pid', 'tag', 'gid'])

        return [pa]

    def create_solver(self):
        print("Create our own solver.")
        kernel = Gaussian(dim=2)

        integrator = EPECIntegrator(fluid=WCSPHStep())

        dt = 5e-6
        tf = 0.0076
        solver = Solver(kernel=kernel, dim=2, integrator=integrator,
                        dt=dt, tf=tf, adaptive_timestep=True,
                        cfl=0.3, n_damp=50,
                        output_at_times=[0.0008, 0.0038])

        return solver

    def create_equations(self):
        print("Create our own equations.")
        equations = [
            Group(
                equations=[
                    TaitEOS(
                        dest='fluid', sources=None, rho0=self.ro,
                        c0=self.co, gamma=7.0
                    ),
                ],
                real=False
            ),
            Group(equations=[
                ContinuityEquation(dest='fluid',  sources=['fluid']),

                MomentumEquation(
                    dest='fluid', sources=['fluid'],
                    alpha=self.alpha, beta=0.0, c0=self.co
                ),

                XSPHCorrection(dest='fluid', sources=['fluid']),

            ]),
        ]
        return equations


if __name__ == '__main__':
    app = EllipticalDrop()
    app.run()
    app.post_process(app.info_filename)



"""Three cubes of different density falling into a static tank of
water in 3d.(20 minutes)

An example to illustrate 3d pysph framework
"""
from __future__ import print_function
import numpy as np

from pysph.base.utils import (get_particle_array_wcsph,
                              get_particle_array_rigid_body)
# PySPH base and carray imports
from pysph.base.kernels import CubicSpline

from pysph.solver.solver import Solver
from pysph.sph.integrator import EPECIntegrator
from pysph.sph.integrator_step import WCSPHStep

from pysph.sph.equation import Group
from pysph.sph.basic_equations import (XSPHCorrection, ContinuityEquation)
from pysph.sph.wc.basic import TaitEOSHGCorrection, MomentumEquation
from pysph.solver.application import Application
from pysph.sph.rigid_body import (
    BodyForce, RigidBodyCollision, RigidBodyMoments,
    RigidBodyMotion, AkinciRigidFluidCoupling, RK2StepRigidBody)


def get_3d_dam(length=10, height=15, depth=10, dx=0.1, layers=2):
    _x = np.arange(0, length, dx)
    _y = np.arange(0, height, dx)
    _z = np.arange(0, depth, dx)

    x, y, z = np.meshgrid(_x, _y, _z)
    x, y, z = x.ravel(), y.ravel(), z.ravel()

    # get particles inside the tank
    tmp = layers - 1
    cond_1 = (x > tmp * dx) & (x < _x[-1] - tmp * dx) & (y > tmp * dx)

    cond_2 = (z > tmp * dx) & (z < z[-1] - tmp * dx)

    cond = cond_1 & cond_2
    # exclude inside particles
    x, y, z = x[~cond], y[~cond], z[~cond]

    return x, y, z


def get_3d_block(length=10, height=15, depth=10, dx=0.1):
    x = np.arange(0, length, dx)
    y = np.arange(0, height, dx)
    z = np.arange(0, depth, dx)

    x, y, z = np.meshgrid(x, y, z)
    x, y, z = x.ravel(), y.ravel(), z.ravel()
    return x, y, z


def get_fluid_and_dam_geometry_3d(d_l, d_h, d_d, f_l, f_h, f_d, d_layers, d_dx,
                                  f_dx, fluid_left_extreme=None,
                                  tank_outside=False):
    xd, yd, zd = get_3d_dam(d_l, d_h, d_d, d_dx, d_layers)
    xf, yf, zf = get_3d_block(f_l, f_h, f_d, f_dx)

    if fluid_left_extreme:
        x_trans, y_trans, z_trans = fluid_left_extreme
        xf += x_trans
        yf += y_trans
        zf += z_trans

    else:
        xf += 2 * d_dx
        yf += 2 * d_dx
        zf += 2 * d_dx

    return xd, yd, zd, xf, yf, zf


def get_sphere(centre=[0, 0, 0], radius=1, dx=0.1):
    x = np.arange(0, radius * 2, dx)
    y = np.arange(0, radius * 2, dx)
    z = np.arange(0, radius * 2, dx)

    x, y, z = np.meshgrid(x, y, z)
    x, y, z = x.ravel(), y.ravel(), z.ravel()

    cond = ((x - radius)**2 + (y - radius)**2) + (z - radius)**2 <= radius**2

    x, y, z = x[cond], y[cond], z[cond]

    x_trans = centre[0] - radius
    y_trans = centre[1] - radius
    z_trans = centre[2] - radius

    x = x + x_trans
    y = y + y_trans
    z = z + z_trans

    return x, y, z


class RigidFluidCoupling(Application):
    def initialize(self):
        self._spacing = 4
        self.spacing = self._spacing * 1e-3
        self.dx = self.spacing
        self.hdx = 1.2
        self.ro = 1000
        self.solid_rho = 800
        self.m = 1000 * self.dx * self.dx * self.dx
        self.co = 2 * np.sqrt(2 * 9.81 * 150 * 1e-3)
        self.alpha = 0.1

    def create_particles(self):
        # get coordinates of tank and fluid
        tank_len = 150
        tank_hei = 150
        tank_dep = 150
        layers = 2

        flu_len = 150 - 2 * layers * self._spacing
        flu_hei = 52
        flu_dep = 150 - 2 * layers * self._spacing

        xt, yt, zt, xf, yf, zf = get_fluid_and_dam_geometry_3d(
            d_l=tank_len, d_h=tank_hei, d_d=tank_dep, f_l=flu_len, f_h=flu_hei,
            f_d=flu_dep, d_layers=2, d_dx=self._spacing, f_dx=self._spacing)
        # scale it to mm
        xt, yt, zt, xf, yf, zf = (xt * 1e-3, yt * 1e-3, zt * 1e-3, xf * 1e-3,
                                  yf * 1e-3, zf * 1e-3)

        # get coordinates of cube
        xc, yc, zc = get_3d_block(20, 20, 20, self._spacing/2.)
        xc1, yc1, zc1 = ((xc + 60) * 1e-3, (yc + 120) * 1e-3, (zc + 70) * 1e-3)
        xc2, yc2, zc2 = ((xc + 4 * self._spacing) * 1e-3, (yc + 120) * 1e-3,
                         (zc + 70) * 1e-3)
        xc3, yc3, zc3 = ((xc + 100) * 1e-3, (yc + 120) * 1e-3,
                         (zc + 70) * 1e-3)

        xc, yc, zc = (np.concatenate((xc1, xc2, xc3)), np.concatenate(
            (yc1, yc2, yc3)), np.concatenate((zc1, zc2, zc3)))

        # Create particle array for fluid
        m = self.ro * self.spacing * self.spacing * self.spacing
        rho = self.ro
        h = self.hdx * self.spacing
        fluid = get_particle_array_wcsph(x=xf, y=yf, z=zf, h=h, m=m, rho=rho,
                                         name="fluid")

        # Create particle array for tank
        m = 1000 * self.spacing**3
        rho = 1000
        rad_s = self.spacing / 2.
        h = self.hdx * self.spacing
        V = self.spacing**3
        tank = get_particle_array_wcsph(x=xt, y=yt, z=zt, h=h, m=m, rho=rho,
                                        rad_s=rad_s, V=V, name="tank")
        for name in ['fx', 'fy', 'fz']:
            tank.add_property(name)

        # Create particle array for cube
        h = self.hdx * self.spacing/2.

        # assign  density of three spheres
        rho1 = np.ones_like(xc1) * 2000
        rho2 = np.ones_like(xc1) * 800
        rho3 = np.ones_like(xc1) * 500
        rho = np.concatenate((rho1, rho2, rho3))

        # assign body id's
        body1 = np.ones_like(xc1, dtype=int) * 0
        body2 = np.ones_like(xc1, dtype=int) * 1
        body3 = np.ones_like(xc1, dtype=int) * 2
        body = np.concatenate((body1, body2, body3))

        m = rho * (self.spacing/2.)**3
        rad_s = self.spacing / 4.
        V = (self.spacing/2.)**3
        cs = 0.0
        cube = get_particle_array_rigid_body(x=xc, y=yc, z=zc, h=h, m=m,
                                             rho=rho, rad_s=rad_s, V=V, cs=cs,
                                             body_id=body,
                                             name="cube")

        print(
            fluid.get_number_of_particles(),
            tank.get_number_of_particles(),
            cube.get_number_of_particles(), )
        return [fluid, tank, cube]

    def create_solver(self):
        kernel = CubicSpline(dim=3)

        integrator = EPECIntegrator(fluid=WCSPHStep(), tank=WCSPHStep(),
                                    cube=RK2StepRigidBody())

        # dt = 0.125 * self.dx * self.hdx / (self.co * 1.1) / 2.
        dt = 1e-4
        print("DT: %s" % dt)
        tf = 0.6
        solver = Solver(
            kernel=kernel,
            dim=3,
            integrator=integrator,
            dt=dt,
            tf=tf,
            adaptive_timestep=False, )

        return solver

    def create_equations(self):
        equations = [
            Group(equations=[
                BodyForce(dest='cube', sources=None, gy=-9.81),
            ], real=False),
            Group(equations=[
                ContinuityEquation(dest='fluid',
                                   sources=['fluid', 'tank', 'cube']),
                ContinuityEquation(dest='tank',
                                   sources=['tank', 'fluid', 'cube'])
            ]),

            # Tait equation of state
            Group(equations=[
                TaitEOSHGCorrection(dest='fluid', sources=None, rho0=self.ro,
                                    c0=self.co, gamma=7.0),
                TaitEOSHGCorrection(dest='tank', sources=None, rho0=self.ro,
                                    c0=self.co, gamma=7.0),
            ], real=False),
            Group(equations=[
                MomentumEquation(dest='fluid', sources=['fluid', 'tank'],
                                 alpha=self.alpha, beta=0.0, c0=self.co,
                                 gy=-9.81),
                AkinciRigidFluidCoupling(dest='fluid', sources=['cube']),
                XSPHCorrection(dest='fluid', sources=['fluid', 'tank']),
            ]),
            Group(equations=[
                RigidBodyCollision(dest='cube', sources=['tank', 'cube'],
                                   kn=1e5)
            ]),
            Group(equations=[RigidBodyMoments(dest='cube', sources=None)]),
            Group(equations=[RigidBodyMotion(dest='cube', sources=None)]),
        ]
        return equations


if __name__ == '__main__':
    app = RigidFluidCoupling()
    app.run()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from pysph.base.gpu_nnps_base cimport *


cdef class OctreeGPUNNPS(GPUNNPS):
    cdef NNPSParticleArrayWrapper src, dst # Current source and destination.

    cdef public list pids
    cdef public list pid_keys
    cdef public list cids
    cdef public list cid_to_idx
    cdef public list max_cid
    cdef public object dst_to_src
    cdef object overflow_cid_to_idx
    cdef object curr_cid
    cdef object max_cid_src

    cdef object helper
    cdef object radix_sort
    cdef object make_vec

    cdef public bint allow_sort
    cdef bint dst_src

    cdef public object neighbor_cid_counts
    cdef public object neighbor_cids
    cdef public list octrees
    cdef public bint use_elementwise
    cdef public bint use_partitions
    cdef public object leaf_size

    cpdef _bin(self, int pa_index)

    cpdef _refresh(self)

    cdef void find_neighbor_lengths(self, nbr_lengths)
    cdef void find_nearest_neighbors_gpu(self, nbrs, start_indices)

    cpdef get_kernel_args(self, c_type)



from starcluster.clustersetup import DefaultClusterSetup
from starcluster.logger import log


class PySPHInstallerBase(DefaultClusterSetup):
    PYSPH_PROFILE = "/etc/profile.d/pysph.sh"
    PYSPH_HOSTS = "/home/pysph/PYSPH_HOSTS"
    PYSPH_USER = "pysph"

    def _create_env(self, master):
        master.ssh.execute(
            r"""
            echo $HOME
            if [ ! -d ~/pysph_env ]; then
            mkdir ~/pysph_env &&
            virtualenv --system-site-packages ~/pysph_env;
            fi
            """
        )

    def _install_pysph(self, master):
        commands = r"""
          . ~/pysph_env/bin/activate
          if ! python -c "import pysph" &> /dev/null; then
          export USE_TRILINOS=1
          export ZOLTAN_INCLUDE=/usr/include/trilinos
          export ZOLTAN_LIBRARY=/usr/lib/x86_64-linux-gnu
          cd ~ &&
          git clone https://github.com/pypr/pysph &&
          cd pysph &&
          python setup.py install
          fi
          """
        master.ssh.execute(commands)

    def _configure_profile(self, node):
        pysph_profile = node.ssh.remote_file(self.PYSPH_PROFILE, 'w')
        pysph_profile.write("test -e ~/.bashrc && . ~/.bashrc")
        pysph_profile.close()


class PySPHInstaller(PySPHInstallerBase):
    def run(self, nodes, master, user, user_shell, volumes):
        aliases = [n.alias for n in nodes]

        log.info("Configuring PYSPH Profile")
        for node in nodes:
            self.pool.simple_job(self._configure_profile,
                                 (node,))
        self.pool.wait(len(nodes))

        master.ssh.switch_user(self.PYSPH_USER)
        log.info("Creating virtual environment")
        self._create_env(master)
        master.ssh.execute("echo '. ~/pysph_env/bin/activate' > ~/.bashrc")

        log.info("Installing PySPH")
        self._install_pysph(master)

        log.info("Adding nodes to PYSPH hosts file")
        pysph_hosts = master.ssh.remote_file(self.PYSPH_HOSTS, 'w')
        pysph_hosts.write('\n'.join(aliases) + '\n')

    def on_add_node(self, new_node, nodes, master, user, user_shell, volumes):
        log.info("Configuring PYSPH Profile")
        self._configure_profile(new_node)

        master.ssh.switch_user(self.PYSPH_USER)
        log.info("Adding %s to PYSPH hosts file" % new_node.alias)
        pysph_hosts = master.ssh.remote_file(self.PYSPH_HOSTS, 'a')
        pysph_hosts.write(new_node.alias + '\n')
        pysph_hosts.close()

    def on_remove_node(self, remove_node, nodes, master,
                       user, user_shell, volumes):
        master.switch_user(self.PYSPH_USER)
        log.info("Removing %s from PYSPH hosts file" % remove_node.alias)
        master.ssh.remove_lines_from_file(self.PYSPH_HOSTS, remove_node.alias)



"""Cylindrical dam break over a dry bed with particle splitting. (67 min)

The case is described in "Accurate particle splitting for smoothed particle
hydrodynamics in shallow water with shock capturing", R. Vacondio, B. D. Rogers
and P.K. Stansby, Int. J. Numer. Meth. Fluids, 69 (2012), pp. 1377-1410.
DOI: 10.1002/fld.2646

"""
import os

# Numpy
from numpy import (ones_like, zeros, pi, arange, concatenate, sin, cos,
                   loadtxt, savez)

# PySPH base
from pysph.base.kernels import CubicSpline
from pysph.base.utils import get_particle_array_swe as gpa_swe

# PySPH solver and integrator
from pysph.solver.application import Application
from pysph.solver.solver import Solver
from pysph.solver.utils import load

# PySPH equations
from pysph.sph.equation import Group
from pysph.sph.swe.basic import (
        GatherDensityEvalNextIteration, NonDimensionalDensityResidual,
        UpdateSmoothingLength, CheckConvergenceDensityResidual, SWEOS,
        SWEIntegrator, SWEStep, CorrectionFactorVariableSmoothingLength,
        ParticleAcceleration, ParticleSplit, CheckForParticlesToSplit,
        DaughterVelocityEval
        )

# PySPH Evaluator
from pysph.tools.sph_evaluator import SPHEvaluator


# Constants
rho_w = 1000.0
g = 9.81
dim = 2


class CylindricalDamBreakSplit(Application):
    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float, dest="hdx", default=1.5,
            help="h/dx value used in SPH to change the smoothness")
        group.add_argument(
            "--dw0", action="store", type=float, dest="dw0", default=10.0,
            help="Initial depth of the fluid column (m)")
        group.add_argument(
            "--r", action="store", type=float, dest="r", default=500.0,
            help="Initial radius of the fluid column (m)")
        group.add_argument(
            "--n", action="store", type=float, dest="n", default=50,
            help="Number of concentric fluid particle circles (Determines\
            spacing btw particles, dr = r/n)")
        group.add_argument(
            "--coeff_A_split", action="store", type=float,
            dest="coeff_A_max", default=2.0, help="Ratio of area of\
            particle to initial area after which the particle splitting is\
            activated")
        group.add_argument(
            "--coeff_h_split", action="store", type=float, dest="coeff_h_max",
            default=3.0, help="Ratio of smoothing length of particle to\
            initial smoothing length after which the particle splitting is\
            deactivated")

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dw0 = self.options.dw0
        self.r = self.options.r
        self.n = self.options.n
        self.coeff_A_max = self.options.coeff_A_max
        self.coeff_h_max = self.options.coeff_h_max
        self.dr = self.r / self.n

        # Particle splitting activated when A_i > A_max and h_i < h_max
        self.A_max = self.coeff_A_max * (1.56*self.dr**2)
        self.h_max = self.coeff_h_max * self.hdx * self.dr

    def create_particles(self):
        dr = self.dr
        n = self.n
        d = self.dw0
        hdx = self.hdx

        x = zeros(0)
        y = zeros(0)

        # Create circular patch in a radial grid
        rad = 0.0
        for j in range(1, n+1):
                npnts = 4 * j
                dtheta = (2*pi) / npnts

                theta = arange(0, 2*pi-1e-10, dtheta)
                rad = rad + dr

                _x = rad * cos(theta)
                _y = rad * sin(theta)

                x = concatenate((x, _x))
                y = concatenate((y, _y))

        m = ones_like(x) * (1.56*dr*dr) * rho_w * d

        rho = ones_like(x) * rho_w * d
        rho0 = ones_like(x) * rho_w * d

        h = ones_like(x) * hdx * dr
        h0 = ones_like(x) * hdx * dr

        pa = gpa_swe(x=x, y=y, m=m, rho=rho, rho0=rho0, h=h, h0=h0,
                     name='fluid')

        compute_initial_props([pa])
        return [pa]

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = SWEIntegrator(fluid=SWEStep())
        tf = 50.0
        solver = Solver(
            kernel=kernel,
            dim=2,
            integrator=integrator,
            cfl=0.3,
            adaptive_timestep=True,
            output_at_times=(10.0, 30.0, 50.0),
            tf=tf
            )
        return solver

    def create_equations(self):
        equations = [
            Group(
                equations=[
                    Group(
                        equations=[
                            GatherDensityEvalNextIteration(
                                dest='fluid', sources=['fluid']
                                ),
                            ]
                        ),
                    Group(
                        equations=[
                            NonDimensionalDensityResidual(dest='fluid')
                            ]
                        ),
                    Group(
                        equations=[
                            UpdateSmoothingLength(dim=dim, dest='fluid')
                            ], update_nnps=True
                        ),
                    Group(
                        equations=[
                            CheckConvergenceDensityResidual(dest='fluid')
                            ],
                    )], iterate=True, max_iterations=10
            ),
            Group(
                equations=[
                    CorrectionFactorVariableSmoothingLength(
                        dest='fluid', sources=['fluid']),
                    ]
                ),
            Group(
                equations=[
                    DaughterVelocityEval(rhow=rho_w, dest='fluid',
                                         sources=['fluid']),
                    ]
                ),
            Group(
                equations=[
                    SWEOS(dest='fluid'),
                    ]
                ),
            Group(
                equations=[
                    ParticleAcceleration(dim=dim, dest='fluid',
                                         sources=['fluid']),
                    ]
                ),
            Group(
                equations=[
                    CheckForParticlesToSplit(dest='fluid', h_max=self.h_max,
                                             A_max=self.A_max)
                    ],
                ),
            ]
        return equations

    def pre_step(self, solver):
        for pa in self.particles:
            ps = ParticleSplit(pa)
            ps.do_particle_split()
        self.nnps.update()

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        import matplotlib
        matplotlib.use('Agg')
        self._plot_depth_of_dam()

    def _plot_depth_of_dam(self):
        from matplotlib import pyplot as plt
        t_to_plot = [10.0, 30.0, 50.0]
        fname_for_plot = []
        for fname in self.output_files:
            data = load(fname)
            t = data['solver_data']['t']
            if t in t_to_plot:
                fname_for_plot.append(fname)

        num = 0
        # Finite volume solution files of this problem for depth of dam 10 m
        # and radius 1000 m at times 10s, 30s and 50s
        files_dir = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                'files_for_output_comparison'
                )
        finite_vol_soln_files = [
                os.path.join(files_dir, 'cyl_dam_split_t01.csv'),
                os.path.join(files_dir, 'cyl_dam_split_t02.csv'),
                os.path.join(files_dir, 'cyl_dam_split_t03.csv')
                ]
        for fname in fname_for_plot:
            data = load(fname)
            fluid = data['arrays']['fluid']
            t = data['solver_data']['t']

            cond = abs(fluid.y) < 1e-2
            x = fluid.x[cond]
            y = fluid.y[cond]

            if len(x) == 0:
                return

            x_max = max(x)
            dw = fluid.dw[cond]
            rho = fluid.rho[cond]
            u = fluid.u[cond]
            v = fluid.v[cond]
            au = fluid.au[cond]
            av = fluid.av[cond]

            plt.clf()
            if (
                self.dw0 == 10.0 and
                self.r == 500.0 and
                (t in [10.0, 30.0, 50.0])
            ):
                file_num = [10.0, 30.0, 50.0].index(t)
                x_fv, dw_fv = loadtxt(
                    finite_vol_soln_files[file_num], delimiter=',',
                    unpack=True)
                x_fv_idx_sort = x_fv.argsort()
                x_cen_fv = 1500.0
                x_fv = x_fv[x_fv_idx_sort] - x_cen_fv
                dw_fv = dw_fv[x_fv_idx_sort]
                plt.plot(x_fv, dw_fv, 'r-', label='Finite Volume Solution')

                fname_res = os.path.join(
                    self.output_dir, 'results%d.npz' % (num+1))
                savez(
                    fname_res, t=t_to_plot[num], x=x, x_fv=x_fv, y=y,
                    dw=dw, dw_fv=dw_fv, u=u, v=v, rho=rho, au=au, av=av)

            plt.plot(x, dw, 'bo', label='SWE-SPH', markersize=2)
            plt.ylim(0, 1.5*self.dw0)
            plt.xlim(-1.5*x_max, 1.5*x_max)
            plt.legend()
            plt.xlabel('x (m)')
            plt.ylabel('dw (m)')
            plt.title('t = %0.1fs' % t_to_plot[num])
            fig = os.path.join(self.output_dir, "depth%d.png" % (num+1))
            plt.savefig(fig, dpi=300)
            num += 1


def compute_initial_props(particles):
    one_time_equations = [
        Group(
            equations=[
                SWEOS(dest='fluid'),
                ],
            )
    ]
    kernel = CubicSpline(dim=2)
    sph_eval = SPHEvaluator(particles, one_time_equations, dim=2,
                            kernel=kernel)
    sph_eval.evaluate()


if __name__ == '__main__':
    app = CylindricalDamBreakSplit()
    app.run()
    app.post_process(app.info_filename)



"""SPHERIC benchmark case 6. (2 hours)

See http://spheric-sph.org/tests/test-6 for more details.
"""

# math
from math import exp

# PySPH imports
from pysph.base.utils import get_particle_array
from pysph.base.kernels import QuinticSpline
from pysph.solver.solver import Solver
from pysph.solver.application import Application

from pysph.sph.integrator_step import TwoStageRigidBodyStep, TransportVelocityStep
from pysph.sph.integrator import Integrator

from pysph.tools import uniform_distribution

# SPH equations for this problem
from pysph.sph.equation import Group, Equation
from pysph.sph.wc.transport_velocity import SummationDensity,\
    StateEquation, MomentumEquationPressureGradient, MomentumEquationViscosity,\
    MomentumEquationArtificialStress, SolidWallPressureBC, SolidWallNoSlipBC,\
    SetWallVelocity

# domain and reference values
Lx = 10.0
Ly = 5.0
Umax = 1.0
c0 = 25.0 * Umax
rho0 = 1.0
p0 = c0 * c0 * rho0

# obstacle dimensions
obstacle_width = 1.0
obstacle_height = 1.0

# Reynolds number and kinematic viscosity
Re = 150
nu = Umax * obstacle_width / Re

# Numerical setup
nx = 50
dx = 0.20 * Lx / nx
nghost_layers = 4
ghost_extent = nghost_layers * dx
hdx = 1.2

# adaptive time steps
h0 = hdx * dx
dt_cfl = 0.25 * h0 / (c0 + Umax)
dt_viscous = 0.125 * h0**2 / nu
dt_force = 1.0

tf = 8.0
dt = 0.8 * min(dt_cfl, dt_viscous, dt_force)


class SPHERICBenchmarkAcceleration(Equation):
    r"""Equation to set the acceleration for the moving square
    benchmark problem.

    We use scipy.optimize to fit the Gaussian:

    .. math::

        a \exp( -\frac{(t-b)^2}{2c^2} ) + d

    to the SPHERIC Motion.dat file. The values for the parameters are

    a = 2.8209512
    b = 0.525652151
    c = 0.14142151
    d = -2.55580905e-08

    Notes:

    This equation must be instantiated with no sources

    """

    def loop(self, d_idx, d_au, t=0.0):
        a = 2.8209512
        b = 0.525652151
        c = 0.14142151
        d = -2.55580905e-08

        # compute the acceleration and set it for the destination
        d_au[d_idx] = a * exp(-(t - b) * (t - b) / (2.0 * c * c)) + d


def _get_interior(x, y):
    indices = []
    for i in range(x.size):
        if ((x[i] > 0.0) and (x[i] < Lx)):
            if ((y[i] > 0.0) and (y[i] < Ly)):
                indices.append(i)

    return indices


def _get_obstacle(x, y):
    indices = []
    for i in range(x.size):
        if ((1.0 <= x[i] <= 2.0) and (2.0 <= y[i] <= 3.0)):
            indices.append(i)

    return indices


class MovingSquare(Application):
    def _setup_particle_properties(self, particles, volume):
        fluid, solid, obstacle = particles

        #### ADD PROPS FOR THE PARTICLES ###

        # volume from number density
        fluid.add_property('V')
        solid.add_property('V')
        obstacle.add_property('V')

        # extrapolated velocities for the fluid
        for name in ['uf', 'vf', 'wf']:
            solid.add_property(name)
            obstacle.add_property(name)

        # dummy velocities for the solid and obstacle
        # required for the no-slip BC
        for name in ['ug', 'vg', 'wg']:
            solid.add_property(name)
            obstacle.add_property(name)

        # advection velocities and accelerations for fluid
        for name in ('uhat', 'vhat', 'what', 'auhat',
                     'avhat', 'awhat', 'au', 'av', 'aw'):
            fluid.add_property(name)

        # kernel summation correction for solids
        solid.add_property('wij')
        obstacle.add_property('wij')

        # initial velocities and positions needed for the obstacle for
        # rigid-body integration
        obstacle.add_property('u0'); obstacle.u0[:] = 0.
        obstacle.add_property('v0'); obstacle.v0[:] = 0.
        obstacle.add_property('w0'); obstacle.w0[:] = 0.

        obstacle.add_property('x0')
        obstacle.add_property('y0')
        obstacle.add_property('z0')

        # imposed accelerations on the solid and obstacle
        solid.add_property('ax')
        solid.add_property('ay')
        solid.add_property('az')

        obstacle.add_property('ax')
        obstacle.add_property('ay')
        obstacle.add_property('az')

        # magnitude of velocity squared
        fluid.add_property('vmag2')

        #### SETUP PARTICLE PROPERTIES ###

        # mass is set to get the reference density of rho0
        fluid.m[:] = volume * rho0
        solid.m[:] = volume * rho0
        obstacle.m[:] = volume * rho0

        # volume is set as dx^2
        fluid.V[:] = 1. / volume
        solid.V[:] = 1. / volume
        obstacle.V[:] = 1. / volume

        # smoothing lengths
        fluid.h[:] = h0
        solid.h[:] = h0
        obstacle.h[:] = h0

        # set the output arrays
        fluid.set_output_arrays(['x', 'y', 'u', 'v', 'vmag2', 'rho', 'p',
                                 'V', 'm', 'h'])

        solid.set_output_arrays(['x', 'y', 'rho', 'p'])
        obstacle.set_output_arrays(['x', 'y', 'u0', 'rho', 'p', 'u'])

        particles = [fluid, solid, obstacle]
        return particles

    def add_user_options(self, group):
        group.add_argument(
            "--hcp", action="store_true", dest="hcp", default=False,
            help="Use hexagonal close packing of particles."
        )

    def create_particles(self):

        hcp = self.options.hcp
        # Initial distribution using Hexagonal close packing of particles
        # create all particles
        global dx
        if hcp:
            x, y, dx, dy, xmin, xmax, ymin, ymax = uniform_distribution.uniform_distribution_hcp2D(
                dx=dx, xmin=-ghost_extent, xmax=Lx + ghost_extent,
                ymin=-ghost_extent, ymax=Ly + ghost_extent)
        else:
            x, y, dx, dy, xmin, xmax, ymin, ymax = uniform_distribution.uniform_distribution_cubic2D(
                dx=dx, xmin=-ghost_extent, xmax=Lx + ghost_extent,
                ymin=-ghost_extent, ymax=Ly + ghost_extent)

        x = x.ravel()
        y = y.ravel()

        # create the basic particle array
        solid = get_particle_array(name='solid', x=x, y=y)

        # now sort out the interior from all particles
        indices = _get_interior(solid.x, solid.y)
        fluid = solid.extract_particles(indices)
        fluid.set_name('fluid')

        solid.remove_particles(indices)

        # sort out the obstacle from the interior
        indices = _get_obstacle(fluid.x, fluid.y)
        obstacle = fluid.extract_particles(indices)
        obstacle.set_name('obstacle')

        fluid.remove_particles(indices)

        print("SPHERIC benchmark 6 :: Re = %d, nfluid = %d, nsolid=%d, nobstacle = %d, dt = %g" % (
            Re, fluid.get_number_of_particles(),
            solid.get_number_of_particles(),
            obstacle.get_number_of_particles(), dt))

        # setup requisite particle properties and initial conditions

        if hcp:
            wij_sum = uniform_distribution.get_number_density_hcp(
                dx, dy, kernel, h0)
            volume = 1. / wij_sum
        else:
            volume = dx * dy

        particles = self._setup_particle_properties(
            [fluid, solid, obstacle], volume=volume
        )

        return particles

    def create_solver(self):
        kernel = QuinticSpline(dim=2)
        integrator = Integrator(fluid=TransportVelocityStep(),
                                obstacle=TwoStageRigidBodyStep())

        solver = Solver(kernel=kernel, dim=2, integrator=integrator,
                        tf=tf, dt=dt, adaptive_timestep=False,
                        output_at_times=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])
        return solver

    def create_equations(self):
        equations = [

            # set the acceleration for the obstacle using the special function
            # mimicking the accelerations provided in the test.
            Group(
                equations=[
                    SPHERICBenchmarkAcceleration(
                        dest='obstacle', sources=None),
                ], real=False
            ),

            # Summation density along with volume summation for the fluid
            # phase. This is done for all local and remote particles. At the
            # end of this group, the fluid phase has the correct density
            # taking into consideration the fluid and solid
            # particles.
            Group(
                equations=[
                    SummationDensity(
                        dest='fluid', sources=[
                            'fluid', 'solid', 'obstacle']),
                ], real=False
            ),

            # Once the fluid density is computed, we can use the EOS to set
            # the fluid pressure. Additionally, the dummy velocity for the
            # channel is set, which is later used in the no-slip wall BC.
            Group(
                equations=[
                    StateEquation(
                        dest='fluid',
                        sources=None,
                        p0=p0,
                        rho0=rho0,
                        b=1.0),
                    SetWallVelocity(dest='solid', sources=['fluid']),
                    SetWallVelocity(dest='obstacle', sources=['fluid']),
                ], real=False
            ),

            # Once the pressure for the fluid phase has been updated, we can
            # extrapolate the pressure to the ghost particles. After this
            # group, the fluid density, pressure and the boundary pressure has
            # been updated and can be used in the integration equations.
            Group(
                equations=[
                    SolidWallPressureBC(
                        dest='obstacle',
                        sources=['fluid'],
                        b=1.0,
                        rho0=rho0,
                        p0=p0),
                    SolidWallPressureBC(
                        dest='solid', sources=['fluid'], b=1.0, rho0=rho0, p0=p0),
                ], real=False
            ),

            # The main accelerations block. The acceleration arrays for the
            # fluid phase are upadted in this stage for all local particles.
            Group(
                equations=[
                    # Pressure gradient terms
                    MomentumEquationPressureGradient(
                        dest='fluid', sources=['fluid', 'solid', 'obstacle'], pb=p0),

                    # fluid viscosity
                    MomentumEquationViscosity(
                        dest='fluid', sources=['fluid'], nu=nu),

                    # No-slip boundary condition. This is effectively a
                    # viscous interaction of the fluid with the ghost
                    # particles.
                    SolidWallNoSlipBC(
                        dest='fluid', sources=['solid', 'obstacle'], nu=nu),

                    # Artificial stress for the fluid phase
                    MomentumEquationArtificialStress(
                        dest='fluid', sources=['fluid']),

                ], real=True
            ),
        ]
        return equations


if __name__ == '__main__':
    app = MovingSquare()
    app.run()



from pysph.sph.equation import Equation
import numpy as np


class GroupParticles(Equation):

    def __init__(self, dest, sources=None, xmin=0.0, xmax=0.0, ymin=0.0,
                 ymax=0.0, zmin=0.0, zmax=0.0, periodic_in_x=False,
                 periodic_in_y=False, periodic_in_z=False):
        self.periodic_in_x = periodic_in_x
        self.periodic_in_y = periodic_in_y
        self.periodic_in_z = periodic_in_z
        self.xlen = abs(xmax - xmin)
        self.xmin = xmin
        self.xmax = xmax
        self.ylen = abs(ymax - ymin)
        self.ymin = ymin
        self.ymax = ymax
        self.zlen = abs(zmax - zmin)
        self.zmin = zmin
        self.zmax = zmax

        super(GroupParticles, self).__init__(dest, sources)

    def loop(self, d_idx, d_cm, d_body_id, d_x, d_y, d_z):
        b = declare('int')
        b = d_body_id[d_idx]
        if self.periodic_in_x:
            if (abs(d_x[d_idx] - d_cm[3 * b]) > (self.xlen / 2.0)):
                if (d_cm[3 * b] > self.xmin + self.xlen / 2.0):
                    d_x[d_idx] += self.xlen
                else:
                    d_x[d_idx] -= self.xlen
        if self.periodic_in_y:
            if (abs(d_y[d_idx] - d_cm[3 * b + 1]) > (self.ylen / 2.0)):
                if (d_cm[3 * b + 1] > self.ymin + self.ylen / 2.0):
                    d_y[d_idx] += self.ylen
                else:
                    d_y[d_idx] -= self.ylen
        if self.periodic_in_z:
            if (abs(d_z[d_idx] - d_cm[3 * b + 2]) > (self.zlen / 2.0)):
                if (d_cm[3 * b + 2] > self.zmin + self.zlen / 2.0):
                    d_z[d_idx] += self.zlen
                else:
                    d_z[d_idx] -= self.zlen
